from typing import List

from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import Query


class DBManagementQueryLibrary:
    @staticmethod
    def get_all_rel_types_query() -> Query:
        # find all relations and return the distinct types

        # language=SQL
        query_str = '''
                MATCH () - [rel] - () RETURN DISTINCT type(rel) AS rel_type
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_all_node_labels_query() -> Query:
        # find all nodes and return the distinct labels

        # language=SQL
        query_str = '''
            MATCH (n) RETURN DISTINCT labels(n) AS label
        '''

        return Query(query_str=query_str)

    @staticmethod
    def get_clear_db_query(db_name) -> Query:
        # language=SQL
        query_str = '''
            CREATE OR REPLACE DATABASE $db_name
            WAIT
        '''

        return Query(query_str=query_str, database="system", template_string_parameters={"db_name": db_name})

    @staticmethod
    def get_delete_relationships_query() -> Query:
        # language=SQL
        # add :auto as flag to indicate the query should run in an implicit transaction
        query_str = ''':auto 
                    MATCH ()-[r]->() 
                    CALL { WITH r 
                    DELETE r 
                    } IN TRANSACTIONS OF 50000 ROWS;
                '''

        return Query(query_str=query_str)

    @staticmethod
    def get_delete_nodes_query() -> Query:
        # language=SQL

        # add :auto as flag to indicate the query should run in an implicit transaction
        query_str = ''':auto
                MATCH (n)
                CALL { WITH n
                DELETE n
                } IN TRANSACTIONS OF 50000 ROWS;
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_replace_db_query(db_name) -> Query:
        # language=SQL
        query_str = '''
                    CREATE OR REPLACE DATABASE $db_name
                    WAIT
                '''

        return Query(query_str=query_str,
                     database="system",
                     template_string_parameters={"db_name": db_name})

    @staticmethod
    def get_constraints_query() -> Query:
        query_str = '''
            SHOW INDEX
        '''

        return Query(query_str=query_str)

    @staticmethod
    def get_constraint_unique_entity_uid_query(node_type=None, entity_key_name="sysId") -> Query:
        if node_type is None:
            node_type = "Entity"
        # language=SQL
        query_str = '''
            CREATE CONSTRAINT $constraint_name IF NOT EXISTS 
            FOR (en:$node_type) REQUIRE en.$entity_key_name IS UNIQUE
            // also set the range index
            OPTIONS {
              indexProvider: 'range-1.0'
            }
        '''
        return Query(query_str=query_str,
                     template_string_parameters={
                         "node_type": node_type,
                         "constraint_name": f"unique_{node_type.lower()}_ids",
                         "entity_key_name": entity_key_name
                     })

    @staticmethod
    def get_set_identifier_index_query(node_type: str, identifier_properties: List[str]):
        properties = f"( {','.join(['en.' + prop for prop in identifier_properties])})"
        # language=SQL
        query_str = '''
                    CREATE CONSTRAINT $constraint_name IF NOT EXISTS 
                    FOR (en:$node_type) REQUIRE $properties IS UNIQUE
                '''
        return Query(query_str=query_str,
                     template_string_parameters={
                         "node_type": node_type,
                         "constraint_name": f"unique_{node_type.lower()}_ids",
                         "properties": properties
                     })

    @staticmethod
    def get_set_unique_log_name_index_query() -> Query:
        # language=SQL
        query_str = '''
            CREATE CONSTRAINT unique_entity_ids IF NOT EXISTS 
            FOR (l:Log) REQUIRE l.name IS UNIQUE
        '''
        return Query(query_str=query_str)

    @staticmethod
    def get_set_sysid_index_query(entity_key_name) -> Query:
        # language=SQL
        query_str = '''
            CREATE RANGE INDEX entity_sys_id_index 
            IF NOT EXISTS FOR (n:Entity) ON (n.$entity_key_name)
        '''
        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_key_name": entity_key_name
                     })

    @staticmethod
    def get_set_activity_index_query() -> Query:
        # language=SQL
        query_str = '''
                CREATE RANGE INDEX activity_index 
                IF NOT EXISTS FOR (a:Activity) ON (a.activity)
            '''
        return Query(query_str=query_str)

    @staticmethod
    def get_set_record_id_as_range_query() -> Query:
        # language=SQL
        query_str = '''
                CREATE RANGE INDEX record_id_range 
                IF NOT EXISTS FOR (r:Record) ON (r.recordId)
        '''
        return Query(query_str=query_str)

    @staticmethod
    def get_set_record_type_range_query() -> Query:
        # language=SQL
        query_str = '''
                CREATE RANGE INDEX record_type_range 
                IF NOT EXISTS FOR (rt:RecordType) ON (rt.type)
        '''
        return Query(query_str=query_str)

    @staticmethod
    def get_node_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all node types and counts
                MATCH (n) 
                WITH n, CASE labels(n)[0]
                    WHEN 'Event' THEN 0
                    WHEN 'Entity' THEN 1
                    WHEN 'Class' THEN 2
                    WHEN 'Log' THEN 3
                    ELSE 4
                END AS sortOrder
                WITH  labels(n)[0] AS label,  count(n) AS numberOfNodes,sortOrder
                RETURN label,  numberOfNodes ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_edge_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all agg rel types and counts
                MATCH () - [r] -> ()
                WHERE r.type is NOT NULL
                WITH r, CASE toUpper(r.type)
                  WHEN 'REL' THEN 0
                  WHEN 'DF' THEN 1
                  ELSE 2
                END as sortOrder
                WITH toUpper(r.type) as aggType, count(r) as aggNumberOfRelations, sortOrder
                RETURN aggType, aggNumberOfRelations ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_aggregated_edge_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all rel types and counts
                MATCH () - [r] -> ()
                // WHERE r.type is  NULL
                WITH r, CASE Type(r)
                  WHEN 'CORR' THEN 0
                  WHEN 'OBSERVED' THEN 1
                  WHEN 'HAS' THEN 2
                  ELSE 3
                END as sortOrder
                WITH Type(r) as type, count(r) as numberOfRelations, sortOrder
                RETURN type, numberOfRelations ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_imported_logs_query() -> Query:
        # language = SQL
        query_str = '''
            MATCH (l:Log)
            RETURN COLLECT(DISTINCT l.name) AS logs
        '''

        return Query(query_str=query_str)



from string import Template
from typing import Union, Optional, List

from ..data_managers.semantic_header import ConstructedNodes, NodeConstructor, Node, \
    RelationConstructor, RecordConstructor, ConstructedRelation
from ..database_managers.db_connection import Query


class SemanticHeaderQueryLibrary:
    @staticmethod
    def get_create_node_by_record_constructor_query(node_constructor: NodeConstructor, merge=True,
                                                    logs: Optional[List[str]] = None) -> Query:
        if merge:
            # for each result node, merge the node if it does not exist yet. Then merge it to the record node.
            # So, even though a Entity may appear in multiple records, it is only created once.
            merge_or_create_node = '''MERGE ($result_node)
                                 MERGE (record) <- [:EXTRACTED_FROM] - ($result_node_name)'''
        else:
            # for each result node, merge the node if it does not exist yet NEITHER does its relation to the record
            # node.
            # This ensures that for each record node, exactly one result node is created.
            # e.g. When creating (:Event) nodes, we create one (:Event) node for each (:EventRecord)
            merge_or_create_node = '''MERGE (record) <- [:EXTRACTED_FROM] - ($result_node)'''

        # get the string for optional properties
        set_property_str = node_constructor.get_set_result_properties_query()

        # in case some labels need to be set, we define the string
        set_label_str = ""
        if node_constructor.set_labels is not None:
            set_label_str = f'SET $set_labels'''

        # in case a correlation needs to be created, we define the string
        infer_corr_str = ""
        # in case multiple correlations can be inferred depending on the record types, we create a string for each
        # inference
        if len(node_constructor.inferred_relationships) > 0:
            infer_corr_str = '''WITH record, $result_node_name'''
            for relationship in node_constructor.inferred_relationships:
                infer_rel_str = '''
                    CALL {WITH record, $result_node_name
                            $record_match
                            MATCH ($event_node) - [:EXTRACTED_FROM] -> (record) <- [:EXTRACTED_FROM] - (
                                $result_node_name)
                                MERGE (event) - [:$relation_type] -> ($result_node_name)}'''
                infer_rel_str = Template(infer_rel_str).safe_substitute({
                    "event_node": relationship.event.get_pattern(name="event"),
                    "record_match": relationship.get_record_type_match(record_name="record"),
                    "relation_type": relationship.relation_type
                })
                infer_corr_str += infer_rel_str
        elif node_constructor.infer_corr_from_event_record:
            # only one correlation is created, create a string for this with the corr type
            # language=SQL
            infer_corr_str = '''
            WITH record, $result_node_name
                                MATCH (event:$event_label) - [:EXTRACTED_FROM] -> (record) <- [:EXTRACTED_FROM] - (
                                $result_node_name)
                                MERGE (event) - [:$corr_type] -> ($result_node_name)'''
        elif node_constructor.infer_corr_from_entity_record:  # TODO update such that only correct events are considered
            # language=SQL
            infer_corr_str = '''
                        WITH record, $result_node_name
                                MATCH (event:$event_label) - [:EXTRACTED_FROM] -> (record) <- [:EXTRACTED_FROM] - (
                                $result_node_name)
                                MERGE (event) - [:$corr_type] -> ($result_node_name)'''

        # in case an observed relations needs to be created, we define the string
        infer_observed_str = ""
        if node_constructor.infer_observed:
            # language=SQL
            infer_observed_str = '''
            WITH record, $result_node_name
                                MATCH (event:$event_label) - [:EXTRACTED_FROM] -> (record) <- [:EXTRACTED_FROM] - (
                                $result_node_name)
                                MERGE (event) <- [:OBSERVED] - ($result_node_name)
                                '''

        # add check to only transform records from the imported logs
        if logs is not None:
            log_str = ",".join([f'"{log}"' for log in logs])
            log_check_str = f"<- [:CONTAINS] - (log:Log) WHERE log.name in [{log_str}]"
        else:
            log_check_str = ""

        # create the overall query where we match the correct record nodes
        # then we create/merge the resulting node and set all labels, properties and inferred relations
        # language=SQL
        query_str = '''
                    CALL apoc.periodic.iterate(
                    'MATCH ($record) $log_check_str
                    $record_matches
                    // order records by elementId, this will determine the order in which events are created
                    // this is important for the temporal ordering of :Event nodes 
                    // when creating DF edges in case the timestamps are similar
                          RETURN record ORDER BY elementId(record)',
                          '$merge_or_create_node
                          $set_label_str
                          $set_property_str
                          $infer_corr_str
                          $infer_observed_str', {batchSize:$batch_size})
                    '''

        query_str = Template(query_str).safe_substitute({
            "set_label_str": set_label_str,
            "set_property_str": set_property_str,
            "infer_corr_str": infer_corr_str,
            "infer_observed_str": infer_observed_str,
            "merge_or_create_node": merge_or_create_node,
            "log_check_str": log_check_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "record": node_constructor.get_prevalent_record_pattern(node_name="record"),
                         "record_matches": node_constructor.get_prevalent_match_record_pattern(node_name="record"),
                         "record_name": "record",
                         "result_node": node_constructor.result.get_pattern(),
                         "result_node_name": node_constructor.result.get_name(),
                         "set_result_properties": node_constructor.get_set_result_properties_query(),
                         "set_labels": node_constructor.get_set_result_labels_query(),
                         "corr_type": node_constructor.corr_type,
                         "event_label": node_constructor.event_label
                     })

    @staticmethod
    def get_associated_record_types_query(logs):
        log_str = ",".join([f'"{log}"' for log in logs])
        log_str = f"[{log_str}]"

        # request all associated record types for specific logs
        query_str = '''
            MATCH (record:Record) - [:IS_OF_TYPE] -> (record_type:RecordType)
            MATCH (record) <- [:CONTAINS] - (log:Log)
            WHERE log.name in $log_str
            RETURN collect(distinct record_type.type) as labels
        '''

        return Query(query_str=query_str,
                     template_string_parameters={"log_str": log_str})


    @staticmethod
    def get_infer_corr_from_parent_query(relation_constructor, use_from):
        if use_from:
            node = relation_constructor.from_node.get_pattern()
            from_or_to = "FROM"
        else:
            node = relation_constructor.to_node.get_pattern()
            from_or_to = "TO"

        # add correlation to a child node if its parent is correlated to an event
        query_str = '''
            CALL apoc.periodic.iterate('
                MATCH (e:Event) --> ($node) - [:$from_or_to] - (relation:$relation_label_str)
                WHERE NOT EXISTS ((e) - [:CORR] -> (relation))
                RETURN DISTINCT relation, e',
                'MERGE (e) - [:$corr_type] -> (relation)',
                {batchSize:$batch_size}
                )       
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "node": node,
                         "from_or_to": from_or_to,
                         "relation_label_str": relation_constructor.result.get_relation_types_str(),
                         "corr_type": relation_constructor.corr_type
                     })

    @staticmethod
    def get_create_relation_by_relations_query(relation_constructor: RelationConstructor) -> Query:
        if relation_constructor.model_as_node:
            # language=sql
            merge_str = '''
                            MERGE ($from_node_name) -[:FROM] -> (relation:$rel_pattern) - [:TO] -> (
                            $to_node_name)
                            '''
        else:
            merge_str = "MERGE ($from_node_name) -[$rel_pattern] -> ($to_node_name)"

        # language=SQL
        query_str = '''
                CALL apoc.periodic.iterate(
                '$relation_queries                        
                RETURN distinct $from_node_name, $to_node_name',
                '$merge_str
                $set_properties_str',                        
                {batchSize: $batch_size})
            '''

        query_str = Template(query_str).safe_substitute({
            "merge_str": merge_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "relation_queries": relation_constructor.get_relations_query(),
                         "from_node_name": relation_constructor.from_node.get_name(),
                         "to_node_name": relation_constructor.to_node.get_name(),
                         "rel_pattern": relation_constructor.result.get_pattern(),
                         "set_properties_str": relation_constructor.get_set_result_properties_query()
                     })

    @staticmethod
    def get_create_relation_by_nodes_query(relation_constructor: RelationConstructor) -> Query:
        if relation_constructor.model_as_node:
            # language=sql
            merge_str = '''
                            MERGE ($from_node_name) -[:FROM] -> (relation:$rel_pattern) - [:TO] -> (
                            $to_node_name)
                            '''
        else:
            merge_str = "MERGE ($from_node_name) -[$rel_pattern] -> ($to_node_name)"

        # language=SQL
        query_str = '''
                    CALL apoc.periodic.iterate(
                    '$node_queries                        
                    RETURN distinct $from_node_name, $to_node_name',
                    '$merge_str
                    $set_properties_str',                        
                    {batchSize: $batch_size})
                '''

        query_str = Template(query_str).safe_substitute({
            "merge_str": merge_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "node_queries": relation_constructor.get_node_query(),
                         "from_node_name": relation_constructor.from_node.get_name(),
                         "to_node_name": relation_constructor.to_node.get_name(),
                         "rel_pattern": relation_constructor.result.get_pattern(),
                         "set_properties_str": relation_constructor.get_set_result_properties_query()
                     })



    @staticmethod
    def get_create_relation_using_record_query(relation_constructor: RelationConstructor,
                                               logs: Optional[List[str]] = None) -> Query:
        # find events that are related to different entities of which one event also has a reference to the other entity
        # create a relation between these two entities
        if relation_constructor.model_as_node:
            # language=sql
            merge_str = '''
                            MERGE ($from_node_name) -[:FROM] -> (relation:$rel_pattern) - [:TO] -> (
                            $to_node_name)
                            MERGE (relation)  - [:EXTRACTED_FROM] -> (record)
                            '''
        else:
            merge_str = "MERGE ($from_node_name) -[$rel_pattern] -> ($to_node_name)"

        # add check to only transform records from the imported logs
        if logs is not None:
            log_str = ",".join([f'"{log}"' for log in logs])
            log_check_str = f"<- [:CONTAINS] - (log:Log) WHERE log.name in [{log_str}]"
        else:
            log_check_str = ""

        # match all records that are related to the correct record types and in specific logs
        # then match all from and to nodes that are extracted from these records
        # merge the resulting node
        # set the optional properties
        query_str = '''     CALL apoc.periodic.iterate('
                            MATCH ($record) $log_check_str
                            $record_matches
                            RETURN record',
                            '
                            MATCH ($from_node) - [:EXTRACTED_FROM] -> (record)
                            MATCH ($to_node) - [:EXTRACTED_FROM] -> (record)
                            $merge_str
                            $set_properties_str',
                            {batchSize:$batch_size})
                        '''

        query_str = Template(query_str).safe_substitute({
            "merge_str": merge_str,
            "log_check_str": log_check_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "from_node": relation_constructor.from_node.get_pattern(),
                         "from_node_name": relation_constructor.from_node.get_name(),
                         "to_node": relation_constructor.to_node.get_pattern(),
                         "to_node_name": relation_constructor.to_node.get_name(),
                         "record": relation_constructor.get_prevalent_record_pattern(node_name="record"),
                         "record_matches": relation_constructor.get_prevalent_match_record_pattern(node_name="record"),
                         "rel_pattern": relation_constructor.result.get_pattern("relation"),
                         "relation_labels": relation_constructor.result.get_relation_types_str(as_list=True),
                         "set_properties_str": relation_constructor.get_set_result_properties_query("relation")
                     })

    @staticmethod
    def get_add_duration_query_str(add_duration) -> str:
        if add_duration:
            # check the type of the timestamp attributes.
            # DATE, DATETIME, TIME --> create Duration between first and second
            # INT, FLOAT --> save difference between second and first
            add_duration_str = '''
                , CASE 
                    WHEN apoc.meta.cypher.type(first.timestamp) IN ["DATE_TIME", "TIME", "DATE"] 
                        AND apoc.meta.cypher.type(second.timestamp) IN ["DATE_TIME", "TIME", "DATE"] 
                        THEN duration.between(first.timestamp, second.timestamp)
                    WHEN apoc.meta.cypher.type(first.timestamp) IN ["INTEGER", "FLOAT"]
                     AND apoc.meta.cypher.type(second.timestamp) IN ["INTEGER", "FLOAT"]
                     THEN  second.timestamp - first.timestamp
                    ELSE NULL
                END AS duration
            '''
        else:
            # we don't want to add the duration, so we use NULL for duration
            # This ensures us that we can use duration in the rest of the queries
            add_duration_str = '''
                , NULL as duration
            '''

        return add_duration_str

    @staticmethod
    def get_create_directly_follows_query(entity: Union[ConstructedNodes, ConstructedRelation], event_label,
                                          add_duration: bool = False) -> Query:
        # find the specific entities and events with a certain label correlated to that entity
        # order all events by time, order_nr and id grouped by a node n
        # collect the sorted nodes as a list
        # unwind the list from 0 to the one-to-last node
        # find neighbouring nodes and add an edge between

        # language=sql

        if event_label == "CompoundEvent":
            if entity.type == "Resource":
                query_str = '''
                     CALL apoc.periodic.iterate(
                        'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                        CALL {
                                WITH e
                                MATCH (e) - [:CONSISTS_OF] -> (single_event:Event)
                                RETURN id(single_event) as min_id ORDER BY id(single_event)
                                LIMIT 1
                            }
                        WITH n , e as nodes ORDER BY e.timestamp, min_id
                        WITH n , collect (nodes) as nodeList
                        UNWIND range(0,size(nodeList)-2) AS i
                        WITH n , nodeList[i] as first, nodeList[i+1] as second
                        RETURN n, first, second $add_duration_str',
                        'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                         SET df.type = "DF"
                         SET df.entityId = n.sysId
                         SET df.duration = duration
                        ',
                        {batchSize: $batch_size})
                    '''
            else:
                query_str = '''
                                     CALL apoc.periodic.iterate(
                                        'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                                        CALL {
                                                WITH e
                                                MATCH (e) - [:CONSISTS_OF] -> (single_event:Event)
                                                RETURN id(single_event) as min_id ORDER BY id(single_event)
                                                LIMIT 1
                                            }
                                        WITH n , e as nodes ORDER BY e.timestamp, min_id
                                        WITH n , collect (nodes) as nodeList
                                        UNWIND range(0,size(nodeList)-2) AS i
                                        WITH n , nodeList[i] as first, nodeList[i+1] as second
                                        RETURN first, second $add_duration_str',
                                        'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                                         SET df.type = "DF"
                                         SET df.duration = duration
                                        ',
                                        {batchSize: $batch_size})
                                    '''

        else:
            query_str = '''
                             CALL apoc.periodic.iterate(
                                'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                                WITH n , e as nodes ORDER BY e.timestamp, ID(e)
                                WITH n , collect (nodes) as nodeList
                                UNWIND range(0,size(nodeList)-2) AS i
                                WITH n , nodeList[i] as first, nodeList[i+1] as second
                                RETURN first, second $add_duration_str',
                                'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                                 SET df.type = "DF"
                                 SET df.duration = duration
                                ',
                                {batchSize: $batch_size})
                            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_labels_string": entity.get_label_string(),
                         "corr_type_string": entity.get_corr_type_strings(),
                         "event_label": event_label,
                         "df_entity": entity.get_df_label(),
                         "entity_type": entity.type,
                         "add_duration_str": SemanticHeaderQueryLibrary.get_add_duration_query_str(add_duration)
                     })

    @staticmethod
    def get_merge_duplicate_df_entity_query(node: ConstructedNodes) -> Query:

        # language=sql
        query_str = '''
                        MATCH (n1:Event)-[rel:$df_entity {entityType: '$entity_type'}]->(n2:Event)
                        WITH n1, n2, collect(rel) AS rels
                        WHERE size(rels) > 1
                        // only include this and the next line if you want to remove the existing relationships
                        UNWIND rels AS rel 
                        DELETE rel
                        MERGE (n1)
                            -[:$df_entity {entityType: '$entity_type', count:size(rels), type: 'DF'}]->
                              (n2)
                    '''
        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_type": node.type,
                         "df_entity": node.get_df_label()
                     })



from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import Query


class ExporterQueryLibrary:

    @staticmethod
    def get_event_log_query(entity: ConstructedNodes, additional_event_attributes) -> Query:
        query_str = '''
                MATCH (e:Event) - [:CORR] -> (n:$node_label)
                RETURN n.sysId as caseId, e.activity as activity, e.timestamp as timestamp $extra_attributes
                ORDER BY n.ID, e.timestamp
            '''

        attributes_query = ",".join(f"e.{attribute} as {attribute}" for attribute in additional_event_attributes)
        return Query(query_str=query_str,
                     template_string_parameters={
                         "node_label": entity.get_label_string(),
                         "extra_attributes": f", {attributes_query}" if len(additional_event_attributes) > 0 else ""
                     })



from ..cypher_queries.task_identification_ql import TaskIdentifierLibrary as tf_ql
from ..data_managers.semantic_header import ConstructedNodes, SemanticHeader
from ..utilities.performance_handling import Performance
from ..database_managers.db_connection import DatabaseConnection


class TaskIdentification:
    def __init__(self, db_connection, semantic_header, resource: str, case: str):
        self.connection = db_connection
        self.resource: ConstructedNodes = semantic_header.get_entity(resource)
        self.case: ConstructedNodes = semantic_header.get_entity(case)

    @Performance.track("resource")
    def identify_tasks(self):
        self.connection.exec_query(tf_ql.get_combine_df_joint_query,
                                   **{
                                       "resource": self.resource,
                                       "case": self.case
                                   })

        self.connection.exec_query(tf_ql.get_create_task_instances_query,
                                   **{"resource": self.resource})
        self.connection.exec_query(tf_ql.get_split_ti_nodes_create_new_1_query)
        self.connection.exec_query(tf_ql.get_split_ti_nodes_create_new_2_query)
        self.connection.exec_query(tf_ql.get_split_ti_nodes_remove_old_query)
        self.connection.exec_query(tf_ql.get_remove_df_joint_query)
        for entity in [self.resource, self.case]:
            self.connection.exec_query(tf_ql.get_correlate_ti_to_entity_query,
                                       **{"entity": entity})
            self.connection.exec_query(tf_ql.get_lift_df_to_task_instances_query,
                                       **{"entity": entity})

    @Performance.track("resource")
    def aggregate_on_task_variant(self):
        self.connection.exec_query(tf_ql.get_aggregate_task_instances_query,
                                   **{"property": "variant"})
        self.connection.exec_query(tf_ql.get_link_task_instances_to_aggregations_query,
                                   **{"property": "variant"})
        for entity in [self.resource, self.case]:
            self.connection.exec_query(tf_ql.get_lift_df_to_task_aggregations_query,
                                       **{
                                           "property": "variant",
                                           "entity": entity
                                       })



from typing import Optional, List, Union

from ..data_managers.semantic_header import ConstructedNodes, ConstructedRelation, Relationship, SemanticHeader, \
    NodeConstructor
from ..database_managers.db_connection import DatabaseConnection
from ..utilities.performance_handling import Performance
from ..cypher_queries.semantic_header_ql import SemanticHeaderQueryLibrary as sh_ql


class EKGUsingSemanticHeaderBuilder:
    def __init__(self, database_connection: DatabaseConnection,
                 semantic_header: SemanticHeader):
        self.connection = database_connection
        self.semantic_header = semantic_header

    def create_nodes_by_records(self, node_types: Optional[List[str]], logs: Optional[List[str]]) -> None:
        # request all associated record labels of the imported logs
        associated_records = self.get_associated_records(logs)

        for node_constructor in self.semantic_header.get_node_by_record_constructors(node_types):
            # check if node_constructor is subset of associated record labels
            # if so, we need to create nodes for this record
            is_subset = set(node_constructor.prevalent_record.record_types).issubset(set(associated_records))
            if logs is None or is_subset:
                self._create_node_by_record(node_constructor=node_constructor, logs=logs)

    def get_associated_records(self, logs):
        # if imported logs, we can return an empty list
        if logs is None:
            return []
        else:
            # request the associated record labels
            result = self.connection.exec_query(sh_ql.get_associated_record_types_query,
                                                **{
                                                    "logs": logs
                                                })
            return result[0]["labels"]

    @Performance.track("node_constructor")
    def _create_node_by_record(self, node_constructor: NodeConstructor, logs: Optional[List[str]]):

        merge_first = node_constructor.get_merge()

        self.connection.exec_query(sh_ql.get_create_node_by_record_constructor_query,
                                   **{
                                       "node_constructor": node_constructor,
                                       "merge": merge_first,
                                       "logs": logs
                                   })

    def create_nodes_by_relations(self, node_types: Optional[List[str]]) -> None:
        for node_constructors in self.semantic_header.get_nodes_constructed_by_relations(node_types).values():
            for node_constructor in node_constructors:
                self._create_node_by_relation(node_constructor=node_constructor)

    @Performance.track("node_constructor")
    def _create_node_by_relation(self, node_constructor: NodeConstructor):
        pass

    def create_relations_using_records(self, relation_types: Optional[List[str]],
                                       logs: Optional[List[str]] = None) -> None:
        # find events that are related to different entities of which one event also has a reference to the other entity
        # create a relation between these two entities
        relation: ConstructedRelation

        # request all associated record labels of the imported logs
        associated_records = self.get_associated_records(logs)

        for relation_constructor in self.semantic_header.get_relations_constructed_by_record(relation_types):
            # check if node_constructor is subset of associated record labels
            # if so, we need to create nodes for this record
            is_subset = set(relation_constructor.prevalent_record.record_types).issubset(set(associated_records))
            if logs is None or is_subset:
                self._create_relations_using_record(relation_constructor=relation_constructor,
                                                    logs=logs)

    @Performance.track("relation_constructor")
    def _create_relations_using_record(self, relation_constructor, logs: Optional[List[str]] = None):
        self.connection.exec_query(sh_ql.get_create_relation_using_record_query,
                                   **{
                                       "relation_constructor": relation_constructor,
                                       "logs": logs
                                   })
        self._create_corr_from_parents(relation_constructor=relation_constructor)

    def create_relations_using_relations(self, relation_types: Optional[List[str]]) -> None:
        relation: ConstructedRelation
        for relation_constructor in self.semantic_header.get_relations_constructed_by_relations(relation_types):
            self._create_relations_using_relation(relation_constructor=relation_constructor)

    def create_relations_using_nodes(self, relation_types: Optional[List[str]]) -> None:
        for relation_constructor in self.semantic_header.get_relations_constructed_by_nodes(relation_types):
            self._create_relations_using_nodes(relation_constructor=relation_constructor)

    @Performance.track("relation_constructor")
    def _create_relations_using_relation(self, relation_constructor):
        self.connection.exec_query(sh_ql.get_create_relation_by_relations_query,
                                   **{
                                       "relation_constructor": relation_constructor
                                   })
        self._create_corr_from_parents(relation_constructor=relation_constructor)

    @Performance.track("relation_constructor")
    def _create_relations_using_nodes(self, relation_constructor):
        self.connection.exec_query(sh_ql.get_create_relation_by_nodes_query,
                                   **{
                                       "relation_constructor": relation_constructor
                                   })
        self._create_corr_from_parents(relation_constructor=relation_constructor)

    def _create_corr_from_parents(self, relation_constructor):
        if relation_constructor.infer_corr_from_reified_parents:
            for use_from in [True, False]:
                self.connection.exec_query(sh_ql.get_infer_corr_from_parent_query,
                                           **{
                                               "relation_constructor": relation_constructor,
                                               "use_from": use_from
                                           })

    def create_df_edges(self, entity_types: List[str], event_label: str, add_duration: bool = False) -> None:
        """
        Create Directly Follows edges for the given entity types using the semantic header information
        :param entity_types: list of entity types for which DF edge should be created. If None, we try for all entities
        :param event_label: the labels of the event nodes for which we want to create the event_label
        :param add_duration: boolean indicating whether we want to add the duration to the df edges
        :return: None
        """
        entity: ConstructedNodes

        if entity_types is None:
            entity_types = [entity.type for entity in self.semantic_header.nodes] \
                           + [relation.type for relation in self.semantic_header.relations if relation.model_as_node]

        for entity in self.semantic_header.nodes:
            if entity.infer_df and entity.type in entity_types:
                self._create_df_edges_for_entity(entity=entity, event_label=event_label, add_duration=add_duration)

        for relation in self.semantic_header.relations:
            if relation.model_as_node and relation.infer_df and relation.type in entity_types:
                self._create_df_edges_for_entity(entity=relation, event_label=event_label, add_duration=add_duration)

    @Performance.track("entity")
    def _create_df_edges_for_entity(self, entity: Union[ConstructedNodes, ConstructedRelation], event_label,
                                    add_duration):
        self.connection.exec_query(sh_ql.get_create_directly_follows_query,
                                   **{
                                       "entity": entity,
                                       "event_label": event_label,
                                       "add_duration": add_duration
                                   })

    def merge_duplicate_df(self):
        node: ConstructedNodes
        for node in self.semantic_header.nodes:
            if node.merge_duplicate_df:
                self._merge_duplicate_df_for_node(node=node)

    @Performance.track("node")
    def _merge_duplicate_df_for_node(self, node: ConstructedNodes):
        self.connection.exec_query(sh_ql.get_merge_duplicate_df_entity_query, **{"node": node})

    def create_static_nodes_and_relations(self):
        print("No implementation yet")
        pass



from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import Query


class InferenceEngineQueryLibrary:
    @staticmethod
    def get_query_infer_items_propagate_upwards_multiple_levels(entity: ConstructedNodes, is_load=True) -> Query:
        # language=sql
        query_str = '''
                MATCH (f2:Event) - [:CORR] -> (n:$entity)
                MATCH (f2) - [:CORR] ->  (equipment:Equipment)
                MATCH (f2) - [:OBSERVED] -> (a2:Activity) - [:AT] -> (l:Location) - [:PART_OF*0..] -> (k:Location) 
                WITH f2, k, equipment, n
                CALL {WITH f2, k, equipment
                    MATCH (f0:Event) - [:OBSERVED] -> (a0: Activity)
                    MATCH (a0) - [:$operation_type] -> (et:EntityType {name: '$entity'})  
                    MATCH (a0) - [:AT] -> (k)
                    MATCH (f0) - [:CORR] ->  (equipment)
                    WHERE f0.timestamp $comparison f2.timestamp
                    RETURN f0 as f0_first
                    ORDER BY f0.timestamp $order_type
                    LIMIT 1}
                MERGE (f0_first) - [:CORR] -> (n)
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity": entity.type,
                         "operation_type": "LOADS" if is_load else "UNLOADS",
                         "comparison": "<=" if is_load else ">=",
                         "order_type": "DESC" if is_load else ""
                     })

    @staticmethod
    def get_query_infer_items_propagate_downwards_multiple_level_w_batching(entity: ConstructedNodes,
                                                                            relative_position: ConstructedNodes) -> \
            Query:
        # language=sql
        query_str = '''
                MATCH (f2:Event) - [:CORR] -> (bp:$relative_position)
                MATCH (f2) - [:CORR] -> (equipment :Equipment)
                MATCH (f2) - [:OBSERVED] -> (a2:Activity) -[:AT]-> (l:Location) - [:PART_OF*0..] -> (k:Location) 
                // ensure f2 should have operated on the required by checking that the activity operates on that entity
                MATCH (a2) - -> (:EntityType {name: '$entity'}) 
                WITH f2, equipment, k, bp
                CALL {WITH f2, equipment, k
                    MATCH (f0: Event)-[:OBSERVED]->(a0:Activity) - [:LOADS] ->  (:EntityType {name: '$entity'})
                    MATCH (a0) - [:AT] -> (k)
                    MATCH (f0)-[:CORR]->(resource)
                    WHERE f0.timestamp <= f2.timestamp
                    // find the first preceding f0
                    RETURN f0 as f0_first_prec
                    ORDER BY f0.timestamp DESC
                    LIMIT 1
                }
                // only merge when f0_first_prec is actually related to the required entity
                WITH f2, [(f0_first_prec)-[:CORR]->(n:$entity)- [:AT_POS] -> (bp) | n] as related_n
                FOREACH (n in related_n | 
                    MERGE (f2) - [:CORR] -> (n)
                )
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity": entity.type,
                         "relative_position": relative_position.type
                     })

    @staticmethod
    def get_query_infer_items_propagate_downwards_one_level(entity: ConstructedNodes) -> Query:
        # language=sql
        query_str = '''
                        MATCH (f1 :Event) - [:CORR] -> (equipment :Equipment)
                        MATCH (f1) - [:OBSERVED] -> (a1:Activity) -[:AT]-> (l:Location)
                        // ensure f2 should have operated on the required by checking that the activity operates on 
                        that  
                        entity
                        MATCH (a1) - -> (:EntityType {name: '$entity'}) 
                        WITH f1, equipment, l
                        CALL {WITH f1, equipment, l
                            MATCH (f0: Event)-[:OBSERVED]->(a0:Activity) - [:LOADS] -> (:EntityType {name: '$entity'})
                            MATCH (a0) - [:AT] -> (l)
                            MATCH (f0)-[:CORR]->(equipment)
                            WHERE f0.timestamp <= f1.timestamp
                            // find the first preceding f0
                            RETURN f0 as f0_first_prec
                            ORDER BY f0.timestamp DESC
                            LIMIT 1
                        }
                        // only merge when f0_first_prec is actually related to a Box
                        WITH f1, [(f0_first_prec)-[:CORR]->(n:$entity) | n] as related_n
                        FOREACH (n in related_n | 
                            MERGE (f1) - [:CORR] -> (n)
                        )
                        '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity": entity.type
                     })

    @staticmethod
    def get_match_entity_with_batch_position_query(entity: ConstructedNodes, relative_position: ConstructedNodes) -> Query:
        # language=sql
        query_str = '''
                    MATCH (e:Event) - [:CORR] -> (b:Box)
                    MATCH (e) - [:CORR] -> (bp:$relative_position)
                    MERGE (b:Box) - [:AT_POS] -> (bp:$relative_position)
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity": entity.type,
                         "relative_position": relative_position.type
                     })



from string import Template

from ..database_managers.db_connection import Query


class TaskIdentifierLibrary:
    @staticmethod
    def get_combine_df_joint_query(resource, case):
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (e1:Event)-[:$df_resource]->(e2:Event)
                 WHERE (e1)-[:$df_case]->(e2)
                 RETURN e1,e2",
                "WITH e1,e2
                    MERGE (e1)-[:DF_JOINT]->(e2)",
                    {batchSize:$batch_size})
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "df_resource": resource.get_df_label(),
                         "df_case": case.get_df_label()
                     })

    @staticmethod
    def get_create_task_instances_query(resource):
        print("get_create_task_instances_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "CALL {
                    MATCH (e1:Event)-[:DF_JOINT]->() WHERE NOT ()-[:DF_JOINT]->(e1)
                    MATCH ()-[:DF_JOINT]->(e2:Event) WHERE NOT (e2)-[:DF_JOINT]->()
                    MATCH p=(e1)-[:DF_JOINT*]->(e2)
                    RETURN p, e1, e2
                    UNION
                    MATCH (e:Event) WHERE (e)-[:CORR]->(:$resource_node_label)
                    AND NOT ()-[:DF_JOINT]->(e) AND NOT (e)-[:DF_JOINT]->()
                    MATCH p=(e) RETURN p, e AS e1, e AS e2
                 }
                 RETURN [event in nodes(p) | event.activity+'+'+event.lifecycle] AS variant,
                    nodes(p) AS events, e1.timestamp AS start_time, e2.timestamp AS end_time",
                "WITH variant, events, start_time, end_time
                    CREATE (ti:TaskInstance {variant:variant, start_time:start_time, end_time:end_time})
                    WITH ti, events
                    UNWIND events AS e
                        CREATE (e)<-[:CONTAINS]-(ti)",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "resource_node_label": resource.type
                     })

    @staticmethod
    def get_split_ti_nodes_create_new_1_query():
        print("get_split_ti_nodes_create_new_1_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ti:TaskInstance)-[:CONTAINS]->(e:Event) WHERE date(ti.start_time) <> date(ti.end_time)
                 WITH ti, date(e.timestamp) AS date, e ORDER BY e.timestamp
                 WITH DISTINCT ti, date, COLLECT(e) AS events
                 WITH events[0] AS e_start, events[size(events)-1] AS e_end
                 WITH e_start, e_end
                 MATCH p=(e_start)-[:DF_JOINT*]->(e_end)
                 WITH p, e_start AS e1, e_end AS e2
                 RETURN [event in nodes(p) | event.activity+'+'+event.lifecycle] AS variant, 
                    nodes(p) AS events, e1.timestamp AS start_time, e2.timestamp AS end_time",
                "WITH variant, events, start_time, end_time
                    CREATE (ti:TaskInstance {variant:variant, start_time:start_time, end_time:end_time})
                    WITH ti, events
                    UNWIND events AS e
                    CREATE (e)<-[:CONTAINS]-(ti)",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str)

    @staticmethod
    def get_split_ti_nodes_create_new_2_query():
        print("get_split_ti_nodes_create_new_2_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ti:TaskInstance)-[:CONTAINS]->(e:Event) WHERE date(ti.start_time) <> date(ti.end_time)
                 WITH ti, date(e.timestamp) AS date, e ORDER BY e.timestamp
                 WITH DISTINCT ti, date, COLLECT(e) AS events
                 WITH events[0] AS e_start, events[size(events)-1] AS e_end
                 WITH e_start, e_end
                 MATCH (e_start) MATCH (e_end) WHERE e_start = e_end
                 MATCH p=(e_start)
                 WITH p, e_start AS e1, e_end AS e2
                 RETURN [event in nodes(p) | event.activity+'+'+event.lifecycle] AS variant, 
                    nodes(p) AS events, e1.timestamp AS start_time, e2.timestamp AS end_time",
                "WITH variant, events, start_time, end_time
                 CREATE (ti:TaskInstance {variant:variant, start_time:start_time, end_time:end_time})
                 WITH ti, events
                 UNWIND events AS e
                 CREATE (e)<-[:CONTAINS]-(ti)",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str)

    @staticmethod
    def get_split_ti_nodes_remove_old_query():
        print("get_split_ti_nodes_remove_old_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ti:TaskInstance) WHERE date(ti.start_time) <> date(ti.end_time)
                 RETURN ti",
                "WITH ti
                 DETACH DELETE ti",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str)

    @staticmethod
    def get_remove_df_joint_query():
        print("get_remove_df_joint_query")
        # language=sql
        query_str = '''
                MATCH ()-[r:DF_JOINT]-()
                DELETE r
                '''

        return Query(query_str=query_str)

    @staticmethod
    def get_correlate_ti_to_entity_query(entity):
        print("get_correlate_ti_to_entity_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ti:TaskInstance)-[:CONTAINS]->(:Event)-[:CORR]->(n:$entity_node_label)
                 RETURN DISTINCT ti, n",
                "WITH ti, n
                    CREATE (ti)-[:CORR]->(n)",
                {batchSize:$batch_size})
                    '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_node_label": entity.type
                     })

    @staticmethod
    def get_lift_df_to_task_instances_query(entity):
        print("get_lift_df_to_task_instances_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (n:$entity_node_label)
                 MATCH (ti:TaskInstance)-[:CORR]->(n)
                 WITH n, ti AS nodes ORDER BY ti.start_time, ID(ti)
                 WITH n, COLLECT (nodes) as nodeList
                 UNWIND range(0, size(nodeList)-2) AS i
                 RETURN n, nodeList[i] as ti_first, nodeList[i+1] as ti_second",
                "WITH n, ti_first, ti_second
                    MERGE (ti_first)-[df:DF_TI_$entity_node_label]->(ti_second)",
                {batchSize:$batch_size})
                    '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_node_label": entity.type
                     })

    @staticmethod
    def get_aggregate_task_instances_query(property):
        print("get_aggregate_task_instances_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ti:TaskInstance) WHERE ti.$property IS NOT NULL
                 RETURN DISTINCT ti.$property AS id, count(*) AS count",
                 "WITH id, count
                  MERGE (ta:TaskAggregation {Type:'$property', id:id, count:count})",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "property": property
                     })

    @staticmethod
    def get_link_task_instances_to_aggregations_query(property):
        print("get_link_task_instances_to_aggregations_query")
        # language=sql
        query_str = '''
                CALL apoc.periodic.iterate(
                "MATCH (ta:TaskAggregation) WHERE ta.Type = '$property'
                 MATCH (ti:TaskInstance) WHERE ti.$property = ta.id
                 RETURN ta, ti",
                "WITH ta, ti
                 CREATE (ti)-[:OBSERVED]->(ta)",
                {batchSize:$batch_size})
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "property": property
                     })

    @staticmethod
    def get_lift_df_to_task_aggregations_query(property, entity):
        print("get_lift_df_to_task_aggregations_query")
        # language=sql
        query_str = '''
                MATCH (ta1:TaskAggregation)<-[:OBSERVED]-(ti1:TaskInstance)-[df:DF_TI_$entity_node_label]
                    ->(ti2:TaskInstance)-[:OBSERVED]->(ta2:TaskAggregation)
                    WHERE ta1.Type = '$property' AND ta2.Type = '$property'
                MATCH (ti1)-[:CORR]->(n:$entity_node_label)<-[:CORR]-(ti2)
                WITH ta1, count(df) AS df_freq, ta2
                MERGE (ta1)-[rel2:DF_TA_$entity_node_label]->(ta2) ON CREATE SET rel2.count=df_freq
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "property": property,
                         "entity_node_label": entity.type
                     })



import collections
from string import Template
from typing import Optional, List, Dict, Any, Tuple, Union
from typing import Callable  # type hint for function returns

import neo4j

from .custom_exceptions import BatchQueryExecutionError
from .models.query import Query
from ..utilities.configuration import Configuration

QueryResult = Optional[List[Dict[str, Any]]]
QueryMapper = Callable[..., Query]
QueryStr = str


class Driver(object):
    def __init__(self, uri, auth):
        self._driver = neo4j.GraphDatabase.driver(uri=uri, auth=auth, max_connection_lifetime=200)

    def get_session(self, database):
        return self._driver.session(database=database)


class DatabaseConnection:
    def __init__(self, uri: str, db_name: str, user: str, password: str, verbose: bool = False,
                 batch_size: int = 100_000):
        self.db_name = db_name
        self.verbose = verbose
        self.batch_size = batch_size
        self.driver = Driver(uri=uri, auth=(user, password))

    def exec_query(self, query: Union[Query, QueryStr, QueryMapper], **kwargs) -> QueryResult:
        transformed_query = self._transform_query(query, **kwargs)
        return self._dispatch_query(query=transformed_query)

    @staticmethod
    def _transform_query(query: Union[Query, QueryStr, QueryMapper], **kwargs) -> Query:
        if isinstance(query, Query):
            transformed_query = query
        elif isinstance(query, QueryStr):
            transformed_query = Query(query_str=query)
        elif isinstance(query, collections.abc.Callable):
            transformed_query = query(**kwargs)  # run function that returns Query object

            if not isinstance(transformed_query, Query):  # Type checker expects Query return
                raise TypeError(f"Expected function to return Query, got {type(transformed_query)}")
        else:
            raise TypeError(f'Unsupported query type: {type(query)}')

        return transformed_query

    def _prepare_query(self, query: Query) -> Tuple[str, Dict[str, Any], str, bool, bool]:
        """
        Normalizes a Query object by filling in default parameters and checking if batching is required.

        @param query: A Query object containing the query string, parameters, and database name.

        @return: A tuple containing:
            - query_str (str): The Cypher query string.
            - kwargs (Dict[str, Any]): The processed query parameters.
            - db_name (str): The name of the database to run the query against.
            - is_batched (bool): True if the query uses batching (e.g., via apoc.periodic.commit), False otherwise.
            - is_implicit (bool): True if the query is an implicit query as indicated by the flag :auto, False otherwise.
        """

        # Unpack Query Object
        query_str = query.query_string
        kwargs = query.kwargs or {}  # replace None value by an emtpy dictionary
        db_name = query.database

        if "batch_size" not in kwargs:  # override batch_size if NOT already defined
            kwargs["batch_size"] = self.batch_size

        if "limit" not in kwargs:  # override limit if NOT already defined
            kwargs["limit"] = self.batch_size

        is_batched = "apoc.periodic.commit" in query_str or "apoc.periodic.iterate" in query_str

        is_implicit = query_str.strip().lower().startswith(":auto")
        if is_implicit:
            query_str = query_str.replace(":auto", "")

        return query_str, kwargs, db_name, is_batched, is_implicit

    def _dispatch_query(self, query: Query) -> QueryResult:
        query_str, query_kwargs, db_name, is_batched, is_implicit = self._prepare_query(query)

        if is_batched:
            limit = query_kwargs.pop("limit")
            return self._run_batched_query(query_str=query_str,
                                           limit=limit,
                                           db_name=db_name,
                                           **query_kwargs)
        else:
            return self._exec_query(query_str=query_str,
                                    db_name=db_name,
                                    is_implicit=is_implicit,
                                    **query_kwargs)

    def _run_batched_query(self, query_str: str, limit: int, db_name: str, **query_kwargs) -> QueryResult:
        failed_batches = 1
        attempts = 0
        result = None
        while failed_batches > 0 and attempts <= 10:
            result = self._exec_query(
                query_str=query_str,
                db_name=db_name,
                **query_kwargs)
            failed_batches = result[0]['failedBatches']
            query_kwargs["batch_size"] = int(limit / 2)
            query_kwargs["batch_size"] = max(10000, query_kwargs["batch_size"])
            attempts += 1
        if failed_batches > 0:
            if "iterate" in query_str:
                raise BatchQueryExecutionError(f"Maximum attempts reached: {result[0]['errorMessages']}")
            else:
                raise BatchQueryExecutionError(f"Maximum attempts reached: {result[0]['batchErrors']}")

        return result

    def _exec_query(self, query_str: str, db_name: str = None, is_implicit=False, **query_kwargs) -> QueryResult:
        """
        Write a transaction of the query to  the server and return the result
        @param query_str: string, query to be executed
        @param db_name: string, Name of the database
        @return: The result of the query or None
        """

        def run_query(tx: neo4j.Transaction, _query_str: str, **_query_kwargs) -> Tuple[QueryResult, neo4j.ResultSummary]:
            """
                Run the query and return the result of the query
                @param tx: transaction class on which we can perform queries to the database
                @param _query_str: string
                @return: The result of the query or None if there is no result
            """
            # get the results after the query is executed
            _result = tx.run(
                query=_query_str,  # I'm aware that this should be a LiteralString, however I cannot enforce this.
                parameters=_query_kwargs
            )
            _result_records = _result.data()  # obtain dict representation
            _summary = _result.consume()  # exhaust the result

            # or empty list
            return _result_records, _summary

        if self.verbose:
            print(f"Executing query in {'IMPLICIT' if is_implicit else 'EXPLICIT'} mode.")
            print(f"Database: {db_name}")
            print(f"Query:\n{query_str}")
            if query_kwargs:
                print(f"Parameters: {query_kwargs}")

        if db_name is None:
            db_name = self.db_name

        with self.driver.get_session(database=db_name) as session:
            try:  # try to commit the transaction, if the transaction fails, it is rolled back automatically
                if is_implicit:
                    result = session.run(query_str, **query_kwargs).data()
                else:
                    result, summary = session.execute_write(run_query, query_str, **query_kwargs)
                return result
            except Exception as inst:  # let user know the transaction failed and close the connection
                print("Latest transaction was rolled back")
                print(f"This was your latest query: {query_str}")
                print(inst)

    @staticmethod
    def set_up_connection(config: Configuration):
        return DatabaseConnection(db_name=config.user, uri=config.uri, user=config.user,
                                  password=config.password, verbose=config.verbose, batch_size=config.batch_size)



import os
from typing import List, Optional

from ..data_managers.semantic_header import SemanticHeader
from ..data_managers.datastructures import DatasetDescriptions
from ..modules.db_management import DBManagement
from ..modules.ekg_builder_semantic_header import EKGUsingSemanticHeaderBuilder
from ..modules.data_importer import Importer
from ..database_managers.db_connection import DatabaseConnection
from ..utilities.performance_handling import Performance


class OcedPg:

    def __init__(self, database_connection: DatabaseConnection, dataset_descriptions: DatasetDescriptions,
                 semantic_header: SemanticHeader,
                 use_sample: bool = False,
                 use_preprocessed_files: bool = False,
                 store_files: bool = True,
                 import_directory=None):
        """
            This is a Class that acts as a facade for used to extract, load and transform their data using OCED-PG

            :param dataset_descriptions: A specification of how the data sets are structured
            :type dataset_descriptions: DatasetDescriptions
            :param perf: a :class:Performance to keep track of the running time of the EKG construction
            :type perf: Performance
            :param use_sample: boolean indicating whether the DB should be build using a sample as specified in the
            ImportedDataStructures, defaults to False
            :type use_sample: bool

        """
        # classes responsible for executing queries
        self.data_importer = Importer(database_connection=database_connection,
                                      data_structures=dataset_descriptions,
                                      use_sample=use_sample,
                                      use_preprocessed_files=use_preprocessed_files,
                                      semantic_header=semantic_header,
                                      import_directory=import_directory,
                                      store_files=store_files)
        self.ekg_builder = EKGUsingSemanticHeaderBuilder(database_connection=database_connection,
                                                         semantic_header=semantic_header)
        self.db_manager = DBManagement(db_connection=database_connection)
        self.semantic_header = semantic_header
        self.dataset_descriptions = dataset_descriptions

    def load_and_transform(self):
        self.load()
        self.transform()

    # region import events
    def load(self, logs=None):
        """
        Import data, both event data as other data, as specified in the DS files.
        Add record labels as specified in the semantic header
        Args:
            logs: list of logs to be imported
        """

        # only import logs that are not imported yet
        # dataset name is considered to be unique
        already_imported_logs = self.db_manager.get_imported_logs()
        if logs is None: # no logs are predefined
            logs = self.dataset_descriptions.get_files_list()
        to_be_imported_logs = [log for log in logs if log not in already_imported_logs]

        return self.data_importer.import_data(to_be_imported_logs=to_be_imported_logs)

    # endregion
    def transform(self, logs=None):
        """
            Transform the record layer into a semantic layer
        """
        self.create_nodes(logs)
        self.create_relations(logs)

    def create_nodes(self, node_types: Optional[List[str]] = None, logs: Optional[List[str]] = None) -> None:
        """
        Create nodes as specified in the semantic header

        Args:
            node_types: list of nodes that should be created. In case of None, all nodes are
            created as specified in the semantic header
            logs: list of logs that need to be transformed

        """
        self.create_nodes_by_records(node_types=node_types, logs=logs)

    def create_nodes_by_records(self, node_types: Optional[List[str]] = None,
                                logs: Optional[List[str]] = None) -> None:
        """
        Create nodes based on records as specified in the semantic header

        Args:
            node_types: list of nodes that should be created. In case of None, all nodes are
            created as specified in the semantic header
            logs: list of logs that need to be transformed

        """
        if logs is None:
            files = self.dataset_descriptions.get_structure_name_file_mapping()
            for name, files in files.items():
                self.ekg_builder.create_nodes_by_records(node_types, logs=files)
        else:
            self.ekg_builder.create_nodes_by_records(node_types, logs=logs)

    def create_relations(self, relation_types: Optional[List[str]] = None,
                         logs: Optional[List[str]] = None) -> None:
        """
        Create relations as specified in the semantic header

        Args:
            relation_types: list of relations that should be created. In case of None, all relations are
            created as specified in the semantic header
            logs: list of logs that need to be transformed

        """
        if logs is None:
            files = self.dataset_descriptions.get_structure_name_file_mapping()
            for name, files in files.items():
                self.create_relations_using_record(relation_types=relation_types, logs=files)
                self.create_relations_using_relations(relation_types=relation_types)
                self.create_relations_using_nodes(relation_types=relation_types)
        else:
            self.create_relations_using_record(relation_types=relation_types, logs=logs)
            self.create_relations_using_relations(relation_types=relation_types)
            self.create_relations_using_nodes(relation_types=relation_types)

    def create_relations_using_record(self, relation_types: Optional[List[str]] = None,
                                      logs: Optional[List[str]] = None) -> None:
        """
        Create relations using records as specified in the semantic header

        Args:
            relation_types: list of relations that should be created. In case of None, all relations are
            created as specified in the semantic header
            logs: list of logs that need to be transformed

        """
        self.ekg_builder.create_relations_using_records(relation_types=relation_types, logs=logs)

    def create_relations_using_relations(self, relation_types: Optional[List[str]] = None) -> None:
        """
        Create relations using a subgraph as specified in the semantic header

        Args:
            relation_types: list of relations that should be created. In case of None, all relations are
            created as specified in the semantic header

        """
        self.ekg_builder.create_relations_using_relations(relation_types)

    def create_relations_using_nodes(self, relation_types: Optional[List[str]] = None) -> None:
        """
        Create relations using a subgraph as specified in the semantic header

        Args:
            relation_types: list of relations that should be created. In case of None, all relations are
            created as specified in the semantic header

        """
        self.ekg_builder.create_relations_using_nodes(relation_types)

    def create_nodes_by_relations(self, node_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create entities based on relations as specified in the semantic header

        :param node_types: list of entity types for which the entities based on relations should be created. In
        case of None, all entities based on relations are created as specified in the semantic header
        :type node_types: List[str], optional
        :return: None

        """
        self.ekg_builder.create_nodes_by_relations(node_types)

    def create_df_edges(self, entity_types: Optional[List[str]] = None, event_label: str = "Event",
                        add_duration=False) -> None:
        """
        Pass on method to ekg_builder to merge parallel directly follows in between batching events

        :return: None
        """
        self.ekg_builder.create_df_edges(entity_types, event_label, add_duration=add_duration)
        self.ekg_builder.merge_duplicate_df()

    def create_static_nodes_and_relations(self) -> None:
        """
        Pass on method to ekg_builder to create static nodes and relations
        No implementation yet

        :return: None
        """
        self.ekg_builder.create_static_nodes_and_relations()



import os
from string import Template
from typing import Optional

from ..data_managers.semantic_header import SemanticHeader
from ..database_managers.db_connection import DatabaseConnection
from ..data_managers.datastructures import DatasetDescriptions
from ..utilities.performance_handling import Performance
from ..cypher_queries.data_importer_ql import DataImporterQueryLibrary as di_ql
from pathlib import Path
import pandas as pd


def pop_log_name(log):
    # get log name if it is in the log
    if "log" in log.columns:
        log_names = log.pop("log")
        log_names = log_names.unique()
        if len(log_names) > 1:
            # todo make error
            raise Exception("Each file should originate from exactly one log, now there are more logs defined.")
        else:
            log_name = log_names[0]
    else:
        log_name = None
    return log, log_name


class Importer:
    def __init__(self, database_connection: DatabaseConnection,
                 data_structures: DatasetDescriptions,
                 semantic_header: SemanticHeader,
                 import_directory: Optional[str] = None,
                 use_sample: bool = False,
                 use_preprocessed_files: bool = False,
                 store_files: bool = False):
        self.connection = database_connection
        self.structures = data_structures.structures
        self.records = semantic_header.records

        self.load_batch_size = 20000
        self.use_sample = use_sample
        self.use_preprocessed_files = use_preprocessed_files
        self.store_files = store_files

        self._import_directory = import_directory

    def import_data(self, to_be_imported_logs) -> None:
        for structure in self.structures:
            required_labels = structure.get_required_labels(records=self.records)

            # read in all file names that match this structure
            for file_name in structure.file_names:
                # only load files that are not imported yet
                if to_be_imported_logs is None or file_name in to_be_imported_logs:
                    # read and import the nodes
                    df_log = structure.read_data_set(file_name=file_name,
                                                     use_sample=self.use_sample,
                                                     use_preprocessed_file=self.use_preprocessed_files,
                                                     store_preprocessed_file=self.store_files)

                    df_log = structure.determine_optional_labels_in_log(df_log, records=self.records)

                    self._import_nodes_from_data(df_log=df_log,
                                                 file_name=file_name,
                                                 required_labels=required_labels)

                    if structure.has_datetime_attribute():
                        # once all events are imported, we convert the string timestamp to the timestamp as used in
                        # Cypher
                        self._reformat_timestamps(structure=structure, required_labels=required_labels)

                    # TODO: move filtering to pandas dataframe
                    self._filter_nodes(structure=structure,
                                       required_labels=required_labels)  # filter nodes according to the
                    # structure

        # self._clear_import_directory()

    @Performance.track("structure")
    def _reformat_timestamps(self, structure, required_labels):
        datetime_formats = structure.get_datetime_formats()
        for attribute, datetime_format in datetime_formats.items():
            if datetime_format.is_epoch:
                self.connection.exec_query(di_ql.get_convert_epoch_to_timestamp_query,
                                           **{
                                               "required_labels": required_labels,
                                               "attribute": attribute,
                                               "datetime_object": datetime_format
                                           })

            self.connection.exec_query(di_ql.get_make_timestamp_date_query,
                                       **{
                                           "required_labels": required_labels,
                                           "attribute": attribute,
                                           "datetime_object": datetime_format
                                       })

    @Performance.track("structure")
    def _filter_nodes(self, structure, required_labels):
        for exclude in [True, False]:
            attribute_values_pairs_filtered = structure.get_attribute_value_pairs_filtered(exclude=exclude)
            for name, values in attribute_values_pairs_filtered.items():
                self.connection.exec_query(di_ql.get_filter_records_by_property_query,
                                           **{
                                               "prop": name,
                                               "values": values,
                                               "exclude": exclude,
                                               "required_labels": required_labels
                                           })

    @Performance.track("file_name")
    def _import_nodes_from_data(self, df_log, file_name, required_labels):
        grouped_by_optional_labels = df_log.groupby(by="labels")
        mapping_str = self._determine_column_mapping_str(df_log)

        for optional_labels_str, log in grouped_by_optional_labels:
            optional_labels = optional_labels_str.split(":")
            labels = required_labels + optional_labels
            labels = list(set(labels))
            labels.remove("")
            new_file_name = self.determine_new_file_name(file_name, optional_labels_str)
            self.import_log_into_db(file_name=new_file_name, labels=labels, mapping_str=mapping_str, log=log)

    def import_log_into_db(self, file_name, labels, mapping_str, log):
        # Temporary save the file in the import directory
        log, log_name = pop_log_name(log)

        self._save_log_grouped_by_labels(log=log, file_name=file_name)
        # first create the record types and log nodes
        self.connection.exec_query(di_ql.get_create_record_types_and_log_query,
                                   **{
                                       "labels": labels,
                                       "log_name": log_name
                                   })
        # when creating the records nodes, relations between the record types and log nodes are created
        self.connection.exec_query(di_ql.get_create_nodes_by_loading_csv_query,
                                   **{
                                       "file_name": file_name,
                                       "log_name": log_name,
                                       "labels": labels,
                                       "mapping": mapping_str
                                   })

    @staticmethod
    def determine_new_file_name(file_name, optional_labels_str):
        if optional_labels_str == "":
            return file_name
        return file_name[:-4] + "_" + optional_labels_str.replace(":", "_") + ".csv"

    def _save_log_grouped_by_labels(self, log, file_name):
        log = log.drop(columns=["labels"])
        log.to_csv(Path(self.get_import_directory(), file_name), index=False)


    def _clear_import_directory(self):
        folder = Path(self.get_import_directory())
        if os.path.exists(folder):
            for filename in os.listdir(folder):
                file_path = os.path.join(folder, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                except Exception as e:
                    print('Failed to delete %s. Reason: %s' % (file_path, e))

    @staticmethod
    def _determine_column_mapping_str(log):
        mapping = {}
        dtypes = log.dtypes.to_dict()
        for col_name, type in dtypes.items():
            if type == object:
                continue  # default is STRING
            elif pd.api.types.is_integer_dtype(type):
                mapping[col_name] = 'INTEGER'
            elif pd.api.types.is_float_dtype(type):
                mapping[col_name] = 'FLOAT'
            elif pd.api.types.is_bool_dtype(type):
                mapping[col_name] = 'BOOLEAN'
            else:
                raise Exception(f"Type for column {col_name} is not defined")

        template_str = '$col_name:{type:"$type"}'
        mapping_list = [Template(template_str).substitute({"col_name": col_name, "type": type}) for col_name, type in
                        mapping.items()]
        mapping_str = '{' + ','.join(mapping_list) + '}'
        return mapping_str

    def retrieve_import_directory(self):
        result = self.connection.exec_query(di_ql.get_import_directory_query)
        # get the correct value from the result
        self._import_directory = result[0]['directory']

    def get_import_directory(self):
        if self._import_directory is None:
            self.retrieve_import_directory()
        return self._import_directory



from typing import List, Set, Optional, Dict

from tabulate import tabulate

from ..cypher_queries.db_managment_ql import DBManagementQueryLibrary as dbm_ql
from ..utilities.performance_handling import Performance


class DBManagement:
    def __init__(self, db_connection, semantic_header=None):
        self.connection = db_connection
        self.semantic_header = semantic_header

    @Performance.track()
    def clear_db(self, replace=True) -> bool:
        """
        Replace or clear the entire database by a new one

        Args:
            replace: boolean to indicate whether the database is fully replaced

        """
        if replace:
            result = self.connection.exec_query(dbm_ql.get_replace_db_query, **{"db_name": self.connection.db_name})
            if result[0]['state'] == 'CaughtUp' and result[0]['success']:
                return True
            else:
                return False
        else:
            self.connection.exec_query(dbm_ql.get_delete_relationships_query)
            self.connection.exec_query(dbm_ql.get_delete_nodes_query)
            return True

    @Performance.track()
    def set_constraints(self, entity_key_name="sysId") -> None:
        """
        Set constraints in Neo4j instance
        """
        self._set_identifier_constraints(entity_key_name=entity_key_name)
        self.connection.exec_query(dbm_ql.get_set_unique_log_name_index_query)
        self.connection.exec_query(dbm_ql.get_set_activity_index_query)
        self.connection.exec_query(dbm_ql.get_set_record_id_as_range_query)
        self.connection.exec_query(dbm_ql.get_set_record_type_range_query)

    def get_constraints(self, ignore_defaults=True):
        results = self.connection.exec_query(dbm_ql.get_constraints_query)
        constraint_names = [result['name'] for result in results]
        if ignore_defaults:
            constraint_names.remove("index_343aff4e")  # default token lookup index for node labels
            constraint_names.remove("index_f7700477")  # default token lookup index for relationship types
        return constraint_names

    def _set_identifier_constraints(self, entity_key_name="sysId"):
        if self.semantic_header is not None:
            for node in self.semantic_header.nodes:
                # set the unique constraint per entity type node
                node_labels = node.get_labels(as_str=False)
                node_type = node.type

                identifier_properties = node.get_identifier_properties()
                if len(identifier_properties) > 0:
                    self.connection.exec_query(dbm_ql.get_set_identifier_index_query,
                                               **{
                                                   "node_type": node.type,
                                                   "identifier_properties": identifier_properties
                                               })

                elif "Entity" in node_labels:
                    self.connection.exec_query(dbm_ql.get_constraint_unique_entity_uid_query,
                                               **{
                                                   "node_type": node_type,
                                                   "entity_key_name": entity_key_name
                                               })
        else:
            # is semantic header is not defined, we just set sysid as range (instead of uniqueness constraint)
            self.connection.exec_query(dbm_ql.get_set_sysid_index_query,
                                       **{
                                           "entity_key_name": entity_key_name
                                       })

    def get_all_rel_types(self) -> List[str]:
        """
        Get all relationship types that are present in Neo4j instance

        Returns:
            A list of strings with all relationship types present in the Neo4j instance
        """

        # execute the query and store the result
        result = self.connection.exec_query(dbm_ql.get_all_rel_types_query)
        # in case there are no rel types, the result is None
        # return in this case an emtpy list
        if result is None:
            return []
        # store the results in a list
        result = [record["rel_type"] for record in result]
        return result

    def get_all_node_labels(self) -> Set[str]:
        """
        Get all node labels that are present in Neo4j instance

        Returns:
            A list of strings with all node labels present in the Neo4j instance
        """

        # execute the query and store the result
        result = self.connection.exec_query(dbm_ql.get_all_node_labels_query)
        # in case there are no labels, return an empty set
        if result is None:
            return set([])
        # some nodes have multiple labels, which are returned as a list of labels
        # therefore we need to flatten the result and take the set
        result = set([record for sublist in result for record in sublist["label"]])
        return result

    def get_statistics(self) -> List[Dict[str, any]]:
        """
        Get the count of nodes per label and the count of relationships per type

        Returns:
            A list containing dictionaries with the label/relationship and its count
        """

        def make_empty_list_if_none(_list: Optional[List[Dict[str, str]]]):
            if _list is not None:
                return _list
            else:
                return []

        node_count = self.connection.exec_query(dbm_ql.get_node_count_query)
        edge_count = self.connection.exec_query(dbm_ql.get_edge_count_query)
        agg_edge_count = self.connection.exec_query(dbm_ql.get_aggregated_edge_count_query)
        result = \
            make_empty_list_if_none(node_count) + \
            make_empty_list_if_none(edge_count) + \
            make_empty_list_if_none(agg_edge_count)
        return result

    def print_statistics(self) -> None:
        """
        Print the statistics nicely using tabulate
        """
        print(tabulate(self.get_statistics()))

    def get_imported_logs(self) -> List[str]:
        result = self.connection.exec_query(dbm_ql.get_imported_logs_query)
        return result[0]['logs']



from typing import Dict, Optional, List, Union
from string import Template

from ..data_managers.datastructures import DataStructure, DatetimeObject
from ..data_managers.semantic_header import RecordConstructor
from ..database_managers.db_connection import Query


def create_mapping_str(mapping: str) -> str:
    """
    Create the string including the information of the datatypes mapping used when importing records.

    :param mapping: The dtype mapping of the imported records as string
    :return: str containing the mapping in Cypher format
    """
    if mapping == "":
        return ""
    mapping_str = ''',{nullValues: [""], mapping:$mapping}''' if mapping != "" else ""
    mapping_str = Template(mapping_str).safe_substitute({"mapping": mapping})
    return mapping_str


def get_match_record_types_mapping(labels):
    if len(labels) == 0:
        return "MATCH (record:Record)"
    record_types = "\n".join(
        [f'''MATCH (record:Record) - [:IS_OF_TYPE] -> (:RecordType {{type:"{label}"}})''' for label in
         labels])
    return record_types


class DataImporterQueryLibrary:
    @staticmethod
    def get_import_directory_query() -> Query:
        """
        Query that gets the import directory of the current running database

        :return: Query object to get import directory of the current running database
        """

        # language=SQL
        query_str = """
            Call dbms.listConfig() YIELD name, value
            WHERE name='server.directories.import'
            RETURN value as directory
        """

        return Query(query_str=query_str)

    @staticmethod
    def get_create_record_types_and_log_query(labels: List[str], log_name: str = None) -> Query:
        query_str = "\n".join(
            [f'''MERGE (:RecordType {{type:"{label}"}})''' for label in labels])
        query_str += f'''\nMERGE (:Log {{name:"{log_name}"}})''' if log_name is not None else ""

        return Query(query_str=query_str)

    @staticmethod
    def get_create_nodes_by_loading_csv_query(labels: List[str], file_name: str, mapping: str,
                                              log_name: str = None) -> Query:
        """
        Create event nodes for each row in the batch with labels
        The properties of each row are also the property of the node
        @param mapping: The dtype mapping of the imported records as string
        @param file_name: the name of the file to be imported
        @param labels: The labels of the record nodes
        @param log_name: the name of the log to be imported

        @return: Query object to create record nodes by loading csv
        """

        if log_name is not None:
            match_log_str = '''\n MATCH (log:Log {name:$log_name})'''
            create_log_str = '''CREATE (record)<-[:CONTAINS]-(log)'''
        else:
            match_log_str = ""
            create_log_str = ""
            log_name = ""

        match_record_types = "\n".join(
            [f'''MATCH ({label}_record:RecordType {{type:"{label}"}})''' for label in labels])
        match_record_types += match_log_str

        create_records = "\n".join([f'''CREATE (record) - [:IS_OF_TYPE] -> ({label}_record)''' for label in labels])
        create_records += create_log_str

        # language=SQL
        query_str = '''
                    CALL apoc.periodic.iterate('
                        CALL apoc.load.csv("$file_name" $mapping_str) yield map as row return row',
                        '$match_record_types
                        CREATE (record:Record)
                        $create_records
                        SET record += row '
                    , {batchSize:$batch_size, parallel:true, retries: 1, params:{log_name: $log_name}});              
                          
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "file_name": file_name,
                         "mapping_str": create_mapping_str(mapping),
                         "match_record_types": match_record_types,
                         "create_records": create_records
                     },
                     parameters={
                         "log_name": log_name
                     })

    @staticmethod
    def get_make_timestamp_date_query(required_labels: List[str], attribute: str,
                                      datetime_object: DatetimeObject) -> Query:
        """
        Create a query to convert the strings of the timestamp to the datetime as used in Neo4j
        Remove the str_timestamp property

        @param required_labels: the required labels of the just imported nodes
        @param attribute: the name of the attribute that should be converted
        @param datetime_object: the DatetimeObject describing how the attribute should be converted

        @return: Query object to convert the timestamps string into timestamp objects

        """
        # language=SQL
        offset = datetime_object.timezone_offset
        offset = f'{attribute}+"{offset}"' if offset != "" else attribute

        # language=SQL
        query_str = '''
                CALL apoc.periodic.iterate(
                '$match_record_types 
                WHERE record.$attribute IS NOT NULL AND NOT apoc.meta.cypher.isType(record.$attribute, "$date_type")
                WITH record, record.$offset as timezone_dt
                WITH record, datetime(apoc.date.convertFormat(timezone_dt, "$datetime_object_format", 
                    "$datetime_object_convert_to")) as converted
                RETURN record, converted',
                'SET record.$attribute = converted',
                {batchSize:$batch_size, parallel:true})
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "match_record_types": get_match_record_types_mapping(labels=required_labels),
                         "datetime_object_format": datetime_object.format,
                         "datetime_object_convert_to": datetime_object.convert_to,
                         "date_type": datetime_object.get_date_type(),
                         "attribute": attribute,
                         "offset": offset
                     })

    @staticmethod
    def get_convert_epoch_to_timestamp_query(required_labels: List[str], attribute: str,
                                             datetime_object: DatetimeObject) -> Query:
        """
        Create a query to convert epoch timestamp to the datetime as used in Neo4j
        Remove the str_timestamp property

        @param required_labels: the required labels of the just imported nodes
        @param attribute: the name of the attribute that should be converted
        @param datetime_object: the DatetimeObject describing how the attribute should be converted

        @return: Query object to convert the epoch timestamps into timestamp objects

        """

        # language=SQL
        # TODO update method to be in line with the one before
        query_str = '''
                CALL apoc.periodic.iterate(
                '$match_record_types 
                WHERE record.$attribute IS NOT NULL AND NOT apoc.meta.cypher.isType(record.$attribute, $date_type)
                WITH record, record.$attribute as timezone_dt
                WITH record, apoc.date.format(timezone_dt, $unit, $dt_format) as converted
                RETURN record, converted',
                'SET record.$attribute = converted',
                {batchSize:$batch_size, parallel:false, 
                params: {unit: $unit,
                        dt_format: $datetime_object_format,
                        date_type: $date_type}})
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "attribute": attribute,
                         "match_record_types": get_match_record_types_mapping(labels=required_labels)
                     },
                     parameters={
                         "unit": datetime_object.unit,
                         "datetime_object_format": datetime_object.format,
                         "date_type": datetime_object.convert_to.replace("ISO_", "")
                     })


@staticmethod
def get_filter_records_by_property_query(prop: str, values: Optional[List[str]] = None,
                                         exclude: bool = True, required_labels=["Record"]) -> Query:
    """
    Create a query to remove nodes and their relationships if they have (exlude) or have not (include) a certain
    attribute or a certain attribute-value pairs.

    @param prop: the name of the property
    @param values: a list of values that the property should (not) have for being removed
    @param exclude: boolean indicating whether nodes should be removed if they match the criteria (exclude=True)
    or be kept (exclude = False)
    @param required_labels: the labels the nodes should have

    @return: Query object to remove the load status attribute of the just imported nodes

    """

    if values is None:  # match all events that have a specific property
        negation = "NOT" if exclude else ""
        # query to delete all records and its relationship with property
        # language=SQL
        query_str = '''
                    CALL apoc.periodic.iterate(
                    // match all records that match property
                    '$match_record_types 
                    WHERE record.$prop IS $negation NULL
                    RETURN record',
                    // delete record and its relationships
                    'DETACH DELETE record',
                    // pass the query parameters
                    {batchSize:$batch_size})
                    '''
        template_string_parameters = {"prop": prop, "negation": negation}
    else:  # match all events with specific property and value
        negation = "" if exclude else "NOT"
        # match all r and delete them and its relationship
        # language=SQL
        query_str = '''
            CALL apoc.periodic.iterate(
                // match all records that match property
                    '$match_record_types 
                    WHERE $negation record.$prop IN $values
                    RETURN record',
                    // delete record and its relationships
                    'DETACH DELETE record',
                    // pass the query parameters
                    {batchSize:$batch_size, params:{values:$values}})
                    
                '''
        template_string_parameters = {
            "prop": prop,
            "negation": negation,
            "match_record_types": get_match_record_types_mapping(labels=required_labels)
        }

    # execute query
    return Query(query_str=query_str,
                 template_string_parameters=template_string_parameters,
                 parameters={
                     "values": values
                 })



from .. import SemanticHeader
from ..cypher_queries.inference_engine_ql import InferenceEngineQueryLibrary as ie_ql
from ..data_managers.semantic_header import ConstructedNodes
from ..utilities.performance_handling import Performance
from ..database_managers.db_connection import DatabaseConnection


class InferenceEngine:
    def __init__(self, db_connection):
        self.connection = db_connection

    def match_entity_with_batch_position(self, entity_type: str, relative_position_type: str) -> None:
        """
        Infer the batch position of a specific entity

        Args:
            entity_type: The type of the entity
             relative_position_type: The type of the relative position

        Raises:
             ValueError: when the entity has not been defined
        """
        entity = SemanticHeader().get_entity(entity_type)
        relative_position = SemanticHeader().get_entity(relative_position_type)
        if entity is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        if relative_position is None:
            raise ValueError(f"{relative_position} is not defined in semantic header")
        self._match_entity_with_batch_position(entity, relative_position)

    @Performance.track("entity")
    def _match_entity_with_batch_position(self, entity: ConstructedNodes, relative_position: ConstructedNodes):
        self.connection.exec_query(ie_ql.get_match_entity_with_batch_position_query, **{
            "entity": entity,
            "relative_position": relative_position
        })

    # rule B
    def infer_items_propagate_downwards_one_level(self, entity_type: str) -> None:
        """
        Infer items while propagating downwards one level

        Args:
            entity_type: The type of the entity

        Raises:
             ValueError: when the entity has not been defined
        """
        entity = SemanticHeader().get_entity(entity_type)
        if entity_type is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self._infer_items_propagate_downwards_one_level(entity)

    @Performance.track("entity")
    def _infer_items_propagate_downwards_one_level(self, entity: ConstructedNodes):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_downwards_one_level, **{"entity": entity})

    # rule C
    def infer_items_propagate_upwards_multiple_levels(self, entity_type: str, is_load=True) -> None:
        """
        Infer items while propagating upwards multiple levels (Rule C)

        Args:
            entity_type: The type of the entity
            is_load: indicating whether we are inferring upwards to load events (true) or unload events (false)

        Raises:
            ValueError: when the entity has not been defined
        """

        entity = SemanticHeader().get_entity(entity_type)
        if entity is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self._infer_items_propagate_upwards_multiple_levels(entity, is_load)

    @Performance.track("entity")
    def _infer_items_propagate_upwards_multiple_levels(self, entity, is_load=True):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_upwards_multiple_levels,
                                   **{
                                       "entity": entity,
                                       "is_load": is_load
                                   })

    # rule D
    def infer_items_propagate_downwards_multiple_level_w_batching(self, entity_type: str,
                                                                  relative_position_type: str, ) -> None:
        """
        Infer items while propagating downwards multiple levels with batching (rule D)

        Args:
            entity_type: The type of the entity
            relative_position_type: The type of the relative position

        Raises:
            ValueError: when the entity has not been defined in semantic header
        """
        entity = SemanticHeader().get_entity(entity_type)
        relative_position = SemanticHeader().get_entity(relative_position_type)
        if entity_type is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self._infer_items_propagate_downwards_multiple_level_w_batching(entity, relative_position)

    @Performance.track("entity")
    def _infer_items_propagate_downwards_multiple_level_w_batching(self, entity, relative_position):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_downwards_multiple_level_w_batching,
                                   **{"entity": entity, "relative_position": relative_position})



from string import Template
from typing import List, Optional


class Property:
    def __init__(self, attribute: str, value: str, node_name: Optional[str],
                 node_attribute: Optional[str], ref_node: Optional[str], ref_attribute: Optional[str]):
        self.attribute = attribute
        self.value = value
        self.node_name = node_name,
        self.node_attribute = node_attribute,
        self.ref_node = ref_node
        self.ref_attribute = ref_attribute

    @staticmethod
    def from_string(property_description):
        if property_description is None:
            return None
        if ":" in property_description:
            components = property_description.split(":")
        else:
            components = property_description.split("=")
        attribute = components[0]
        value = components[1]
        attribute = attribute.strip()
        value = value.strip()

        ref_node, ref_attribute, node_name, node_attribute = None, None, None, None
        if "." in value:
            components = value.split(".")
            ref_node = components[0]
            ref_attribute = components[1]

        if "." in attribute:
            components = attribute.split(".")
            node_name = components[0]
            node_attribute = components[1]

        return Property(attribute=attribute, value=value, node_name=node_name,
                        node_attribute=node_attribute, ref_node=ref_node, ref_attribute=ref_attribute)

    def get_pattern(self, is_set=False, name=None):
        if not is_set:
            return f"{self.attribute}: {self.value}"
        else:
            if name is None:
                return f"{self.attribute} = {self.value}"
            else:
                return f"{name}.{self.attribute} = COALESCE({name}.{self.attribute}, {self.value})"

    def __repr__(self):
        return self.get_pattern()


class Properties:
    def __init__(self, identifier_properties: List[Property],
                 required_properties: List[Property],
                 optional_properties: List[Property]):
        self.identifier_properties = identifier_properties
        self.required_properties = required_properties
        self.optional_properties = optional_properties

    def get_required_properties(self):
        return self.identifier_properties + self.required_properties

    def get_identifier_property_attributes(self):
        return [prop.attribute for prop in self.identifier_properties]

    @staticmethod
    def from_string(properties_str: str):
        """
        Interpret properties as a string and return it as a Properties object.
        :param properties_str:
        :return:
        """
        if properties_str is None:
            return None

        properties_str = properties_str.replace("}", "")
        properties = properties_str.split(",")

        # identify whether properties are identifier or other
        other_properties = [prop for prop in properties if "IDENTIFIER" not in prop]
        identifier_properties = [prop.replace("IDENTIFIER", "") for prop in properties if "IDENTIFIER" in prop]

        required_properties = [prop for prop in other_properties if "OPTIONAL" not in prop]
        optional_properties = [prop.replace("OPTIONAL", "") for prop in other_properties if "OPTIONAL" in prop]

        identifier_properties = [Property.from_string(prop) for prop in identifier_properties]
        required_properties = [Property.from_string(prop) for prop in required_properties]
        optional_properties = [Property.from_string(prop) for prop in optional_properties]

        return Properties(
            identifier_properties=identifier_properties,
            required_properties=required_properties,
            optional_properties=optional_properties
        )

    def get_string(self, with_brackets=False, with_optional=False):
        '''
        Return a string representation of the properties
        :param with_brackets:
        :param with_optional:
        :return:
        '''
        properties = [req_prop.get_pattern() for req_prop in self.get_required_properties()]
        if with_optional:
            properties += [f"OPTIONAL {prop.get_pattern()}" for prop in self.optional_properties]

        properties = ", ".join(properties)
        if with_brackets:
            property_string = "{$properties}"
            properties = Template(property_string).substitute(properties=properties)
        return properties

    def get_set_optional_properties_query(self, name):
        if len(self.optional_properties) == 0:
            return None
        return ",".join(
            [prop.get_pattern(is_set=True, name=name) for prop in self.optional_properties])

    def get_idt_properties_query(self, node_name):
        if self.identifier_properties is None:
            return None
        return ",".join(
            [f"{node_name}.{prop.attribute} as {prop.attribute}" for prop in self.identifier_properties])

    def has_required_properties(self, with_optional=False):
        if with_optional:
            return len(self.get_required_properties()) + len(self.optional_properties) > 0
        return len(self.get_required_properties()) > 0



from string import Template
from typing import Optional, Dict


class Query:
    def __init__(self, query_str: str, database: str = None, parameters: Optional[Dict[str, any]] = None,
                 template_string_parameters: Optional[Dict[str, any]] = None):
        if template_string_parameters is not None:
            self.query_string = Template(query_str).safe_substitute(template_string_parameters)
        else:
            self.query_string = query_str
        self.kwargs = parameters
        self.database = database



# remote credentials --> STORE REMOTE CREDENTIALS SOMEWHERE DIFFERENT (not pushed to github)
from .authentication import Credentials

remote = Credentials(
    uri="[your_uri]",
    user="neo4j",
    password="[your_password]"
)


import yaml
from pathlib import Path


class Configuration:
    def __init__(self, semantic_header_path, dataset_description_path, db_name, uri, user, password,
                 verbose, batch_size, use_sample, use_preprocessed_files, import_directory=None):
        self.semantic_header_path = semantic_header_path
        self.dataset_description_path = dataset_description_path
        self.import_directory = import_directory
        self.db_name = db_name
        self.uri = uri
        self.user = user
        self.password = password
        self.verbose = verbose
        self.batch_size = batch_size
        self.use_sample = use_sample
        self.use_preprocessed_files = use_preprocessed_files

    @staticmethod
    def init_conf_with_config_file(path=Path("config.yaml")):
        config = yaml.safe_load(open(path))
        _semantic_header_path = Path(config['semantic_header_path'])
        _dataset_description_path = Path(config['dataset_description_path'])
        if "import_directory" in config:
            _import_directory = Path(config['import_directory'])
        else:
            _import_directory = None
        _db_name = config["db_name"]
        _uri = config["uri"]
        _user = config["user"]
        _password = config["password"]
        _verbose = config["verbose"]
        _batch_size = config["batch_size"]
        _use_sample = config["use_sample"]
        _use_preprocessed_files = config["use_preprocessed_files"]

        return Configuration(semantic_header_path=_semantic_header_path,
                             dataset_description_path=_dataset_description_path,
                             import_directory=_import_directory,
                             db_name=_db_name,
                             uri=_uri,
                             user=_user,
                             password=_password,
                             verbose=_verbose,
                             batch_size=_batch_size,
                             use_sample=_use_sample,
                             use_preprocessed_files=_use_preprocessed_files)



import os
from typing import List, Optional, Dict, Any

import pandas as pd

from ..cypher_queries.exporter_ql import ExporterQueryLibrary as exporter_ql
from ..data_managers.semantic_header import ConstructedNodes, SemanticHeader
from ..database_managers.db_connection import DatabaseConnection


class Exporter:
    def __init__(self, db_connection):
        self.connection = db_connection

    def get_event_log(self, entity_type, additional_event_attributes: Optional[List[str]] = None) -> List[
        Dict[str, Any]]:
        """
        Get an event log extracted from the EKG for a specific entity and return it

        Args:
            entity_type: The type of the entity --> contract states that it exists
            additional_event_attributes: list of different attributes of event that should also be stored in the
            event log

        Returns:
            A list of events with its attributes in the form of a dictionary

        Raises:
             ValueError: when the entity has not been defined
        """

        entity_object = SemanticHeader().get_entity(entity_type)
        if entity_object is None:
            raise ValueError(f"Entity {entity_type} is not defined in the semantic header")

        return self._get_event_log(entity_object, additional_event_attributes)

    def _get_event_log(self, entity: ConstructedNodes, additional_event_attributes: Optional[List[str]] = None) -> \
            Optional[List[Dict[str, Any]]]:
        """
        Get an event log extracted from the EKG for a specific entity and return it

        Args:
            entity: The entity for which we want to extract an event log
            additional_event_attributes: list of different attributes of event that should also be stored in the
            event log

        Returns:
            A list of events with its attributes in the form of a dictionary
        """
        if additional_event_attributes is None:
            additional_event_attributes = []
        return self.connection.exec_query(exporter_ql.get_event_log_query,
                                          **{
                                              "entity": entity,
                                              "additional_event_attributes": additional_event_attributes
                                          })

    def save_event_log(self, entity_type: str, additional_event_attributes: Optional[List[str]] = None) -> None:
        """
        Create an event log extracted from the EKG from a specific entity and store it as a csv file

        Args:
            entity_type: The type of the entity
             additional_event_attributes: list of different attributes of event that should also be stored in the
             event log
        """

        event_log = self.get_event_log(entity_type, additional_event_attributes)
        df = pd.DataFrame(event_log)

        current_file_path = os.path.dirname(__file__)

        dir_path = os.path.join(current_file_path, '../../..', '..', 'data', SemanticHeader().name, 'event_logs')
        file_path = os.path.join(dir_path, f"{entity_type}.csv")
        os.makedirs(dir_path, exist_ok=True)
        df.to_csv(file_path, index=True, index_label="idx")



import json
import os
import re
import warnings
import random
from pathlib import Path

from typing import List, Dict, Any, Optional
from dataclasses import dataclass

import numpy as np
import pandas as pd
from pandas import DataFrame

from .semantic_header import RecordConstructor
from ..utilities.auxiliary_functions import replace_undefined_value, create_list
from ..utilities.configuration import Configuration


@dataclass
class DatetimeObject:
    format: str
    timezone_offset: str
    convert_to: str
    is_epoch: bool
    unit: str

    def get_date_type(self):
        if self.convert_to == "ISO_DATE":
            return "DATE"
        else:
            return "DATE_TIME"

    @staticmethod
    def from_dict(obj: Any) -> 'DatetimeObject':
        if obj is None:
            return None
        _format = obj.get("format")
        _timezone_offset = replace_undefined_value(obj.get("timezone_offset"), "")
        _convert_to = "ISO_DATE"
        if re.search('[hkHK]', _format):
            _convert_to = "ISO_DATE_TIME"

        # _convert_to = str(obj.get("convert_to"))
        _is_epoch = replace_undefined_value(obj.get("is_epoch"), False)
        _unit = obj.get("unit")
        return DatetimeObject(_format, _timezone_offset, _convert_to, _is_epoch, _unit)


@dataclass
class Column:
    name: str
    dtype: str
    nan_values: List[str]
    optional: bool
    range_start: int
    range_end: int

    @staticmethod
    def from_dict(obj: Any) -> Optional['Column']:
        if obj is None:
            return None
        _name = obj.get("name")
        _dtype = obj.get("dtype")
        _nan_values = replace_undefined_value(obj.get("nan_values"), [])
        _optional = replace_undefined_value(obj.get("optional"), False)
        _range_start = obj.get("range_start")
        _range_end = obj.get("range_end")
        return Column(_name, _dtype, _nan_values, _optional, _range_start, _range_end)


@dataclass
class Attribute:
    name: str
    columns: List[Column]
    separator: str
    is_datetime: bool
    is_compound: bool
    optional: bool
    datetime_object: DatetimeObject
    na_rep_value: Any
    na_rep_columns: List[Column]
    filter_exclude_values: List[str]
    filter_include_values: List[str]
    use_filter: bool
    is_primary_key: bool
    is_foreign_key: bool

    @staticmethod
    def from_dict(obj: Any) -> Optional['Attribute']:
        if obj is None:
            return None
        _name = obj.get("name")
        _columns = create_list(Column, obj.get("columns"))
        _is_compound = len(_columns) > 1
        _optional = bool(obj.get("optional"))
        _datetime_object = DatetimeObject.from_dict(obj.get("datetime_object"))
        _is_datetime = _datetime_object is not None
        _na_rep_value = obj.get("na_rep_value")
        _na_rep_columns = create_list(Column, obj.get("na_rep_columns"))
        _separator = obj.get("separator")
        _filter_exclude_values = obj.get("filter_exclude_values")
        _filter_include_values = obj.get("filter_include_values")
        _use_filter = _filter_include_values is not None or _filter_exclude_values is not None  # default value
        _use_filter = replace_undefined_value(obj.get("use_filter"), _use_filter)
        _is_primary_key = replace_undefined_value(obj.get("is_primary_key"), False)
        _is_foreign_key = replace_undefined_value(obj.get("is_foreign_key"), False)
        return Attribute(name=_name, optional=_optional, columns=_columns, separator=_separator,
                         is_compound=_is_compound,
                         is_datetime=_is_datetime, datetime_object=_datetime_object,
                         na_rep_value=_na_rep_value, na_rep_columns=_na_rep_columns,
                         filter_exclude_values=_filter_exclude_values, filter_include_values=_filter_include_values,
                         use_filter=_use_filter, is_primary_key=_is_primary_key, is_foreign_key=_is_foreign_key)


@dataclass
class Sample:
    file_name: str
    use_random_sample: bool
    population_column: str
    size: int
    ids: List[Any]
    between: List[str]
    format: Optional[str]

    @staticmethod
    def from_dict(obj: Any, default_file_name: Optional[str] = None) -> Optional['Sample']:
        if obj is None:
            return None
        # when there is a single file specified, then we can use the default file name
        if default_file_name is not None:
            _file_name = replace_undefined_value(obj.get("file_name"), default_file_name)
        else:
            _file_name = obj.get("file_name")
        _use_random_sample = obj.get("use_random_sample")
        _population_column = obj.get("population_column")
        _size = obj.get("size")
        _ids = obj.get("ids")
        _between = obj.get("between")
        _format = obj.get("datetime_format")

        return Sample(_file_name, _use_random_sample, _population_column, _size, _ids, _between, _format)


class DataStructure:
    def __init__(self, include: bool, name: str, file_directory: str, file_names: List[str],
                 encoding: str, seperator: str, decimal: str,
                 labels: List[str], true_values: List[str], false_values: List[str],
                 add_log: bool, add_index: bool,
                 samples: Dict[str, Sample], attributes: Dict[str, Attribute],
                 split_combined_events: bool):
        self.include = include
        self.name = name
        self.file_directory = file_directory
        self.preprocessed_file_directory = os.path.join(self.file_directory, "preprocessed_files")
        self.file_names = file_names
        self.encoding = encoding
        self.seperator = seperator
        self.decimal = decimal
        self.labels = labels
        if self.labels == ["Record"]:
            self.labels = None
        self.true_values = true_values
        self.false_values = false_values
        self.add_log = add_log
        self.add_index = add_index
        self.samples = samples
        self.attributes = attributes
        self.split_combined_events = split_combined_events
        self.required_labels = None

    def __repr__(self):
        return self.name

    def has_datetime_attribute(self):
        return any([attribute.is_datetime for attribute in self.attributes.values()])
        # return "Event" in self.labels or "EventRecord" in self.labels

    def contains_composed_events(self):
        contains_composed_events = "startTimestamp" in self.get_datetime_formats() \
                                   or "completeTimestamp" in self.get_datetime_formats()
        return self.has_datetime_attribute() and contains_composed_events

    @staticmethod
    def from_dict(obj: Any) -> Optional['DataStructure']:
        if obj is None:
            return None

        _include = replace_undefined_value(obj.get("include"), True)

        if not _include:
            return None

        _name = obj.get("name")
        _path_of_executed_file = os.getcwd()
        _file_directory = os.path.join(os.getcwd(), *obj.get("file_directory").split("\\"))
        _file_names = obj.get("file_names") if obj.get("file_names") is not None else [obj.get("file_name")]
        _encoding = replace_undefined_value(obj.get("encoding"), "utf-8")
        _seperator = replace_undefined_value(obj.get("seperator"), ",")
        _decimal = replace_undefined_value(obj.get("decimal"), ".")
        _labels = obj.get("labels")
        _true_values = obj.get("true_values")
        _false_values = obj.get("false_values")
        _add_log = replace_undefined_value(obj.get("add_log"), False)
        _add_index = replace_undefined_value(obj.get("add_index"), True)

        _samples_obj = obj.get("samples") if obj.get("samples") is not None else obj.get("sample")
        if len(_file_names) == 1:  # single file name is defined
            _samples = create_list(Sample, _samples_obj, _file_names[0])
        else:
            _samples = create_list(Sample, _samples_obj)

        _samples = {sample.file_name: sample for sample in _samples}
        _attributes = create_list(Attribute, obj.get("attributes"))
        _attributes = {attribute.name: attribute for attribute in _attributes}
        _split_combined_events = replace_undefined_value(obj.get("split_combined_events"), False)

        return DataStructure(_include, _name, _file_directory, _file_names, _encoding, _seperator, _decimal,
                             _labels, _true_values, _false_values, _add_log, _add_index,
                             _samples, _attributes, _split_combined_events)

    def get_primary_keys(self):
        return [attribute_name for attribute_name, attribute in self.attributes.items() if attribute.is_primary_key]

    def get_primary_keys_as_attributes(self):
        primary_keys = self.get_primary_keys()
        primary_key_with = [f"n.{primary_key} as {primary_key}" for primary_key in primary_keys]
        primary_key_string = ", ".join(primary_key_with)
        return primary_key_string

    def get_label_string(self):
        return ":".join(self.labels)

    def get_foreign_keys(self):
        return [attribute_name for attribute_name, attribute in self.attributes.items() if attribute.is_foreign_key]

    def get_dtype_dict(self):
        dtypes = {}
        for attribute in self.attributes.values():
            for column in attribute.columns:
                if column.dtype is not None:
                    if column.name not in dtypes:
                        dtypes[column.name] = column.dtype
                    elif column.dtype != dtypes[column.name]:
                        warnings.warn(
                            f"Multiple dtypes ({column.dtype} != {dtypes[column.name]}) "
                            f"defined for {column.name}")
        return dtypes

    def get_required_columns(self):
        required_columns = set()
        for attribute in self.attributes.values():
            # add column names to the required columns
            required_columns.update([x.name for x in attribute.columns])
            required_columns.update([x.name for x in attribute.na_rep_columns])

        return list(required_columns)

    def create_sample(self, file_name, df_log):
        if self.samples is None or file_name not in self.samples:
            warnings.warn(f"No sample population has been defined for {self.name}")
            return df_log

        sample = self.samples[file_name]
        sample_column = sample.population_column
        if sample.use_random_sample:
            random_selection = random.sample(df_log[sample_column].unique().tolist(), k=sample.size)
            df_log = df_log[df_log[sample_column].isin(random_selection)]
        else:
            if sample.ids is not None:
                random_selection = sample.ids
                df_log = df_log[df_log[sample_column].isin(random_selection)]
            elif sample.between is not None:
                start_date = sample.between[0]
                end_date = sample.between[1]
                df_log['date_time_conv'] = pd.to_datetime(df_log[sample_column], format=sample.format)
                df_log = df_log.loc[df_log["date_time_conv"].between(start_date, end_date)]
                df_log = df_log.drop(columns=['date_time_conv'])

        df_log = df_log.reset_index(drop=True)
        return df_log

    @staticmethod
    def replace_nan_values_based_on_na_rep_columns(df_log, attribute):
        if len(attribute.na_rep_columns) != len(attribute.columns):
            raise Exception(
                f"Na_rep_columns does not have the same size as columns for attribute {attribute.name}")
        else:  # they are the same size
            for i, na_rep_column in zip(range(len(attribute.na_rep_columns)), attribute.na_rep_columns):
                attribute_name = f"{attribute.name}_{i}"
                df_log[attribute_name].fillna(df_log[na_rep_column.name], inplace=True)

        return df_log

    @staticmethod
    def replace_nan_values_based_on_na_rep_value(df_log, attribute):
        for i in range(len(attribute.columns)):
            attribute_name = f"{attribute.name}_{i}"
            df_log[attribute_name].fillna(attribute.na_rep_value, inplace=True)

        return df_log

    @staticmethod
    def replace_nan_values_with_unknown(df_log, attribute):
        column: Column
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            if not column.optional:
                try:
                    df_log[attribute_name].fillna("Unknown", inplace=True)
                except:
                    df_log[attribute_name].fillna(-1, inplace=True)
        return df_log

    @staticmethod
    def create_compound_attribute(df_log, attribute):
        compound_column_names = [x.name for x in attribute.columns]
        df_log[attribute.name] = df_log[compound_column_names].apply(
            lambda row: attribute.separator.join([value for value in row.values.astype(str) if
                                                  not (value == 'nan' or value != value)]), axis=1)
        return df_log

    @staticmethod
    def combine_attribute_columns(df_log, attribute):
        compound_attribute_names = [f"{attribute.name}_{i}" for i in range(len(attribute.columns))]
        if attribute.is_compound:
            df_log[f"{attribute.name}_attribute"] = df_log[compound_attribute_names].apply(
                lambda row: attribute.separator.join([value for value in row.values.astype(str) if
                                                      not (value == 'nan' or value != value)]), axis=1)
        else:
            df_log[f"{attribute.name}_attribute"] = df_log[f"{attribute.name}_0"]
        df_log = df_log.drop(columns=compound_attribute_names)
        return df_log

    @staticmethod
    def create_attribute_columns(df_log, attribute):
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            df_log[attribute_name] = df_log[column.name]
            if column.range_start is not None or column.range_end is not None:
                df_log[attribute_name] = df_log[attribute_name].str[column.range_start:column.range_end]
        return df_log

    @staticmethod
    def replace_with_nan(df_log, attribute):
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            for nan_value in column.nan_values:
                df_log[attribute_name] = df_log[attribute_name].replace(nan_value, np.nan, regex=False)
        return df_log

    def preprocess_according_to_attributes(self, df_log):
        # loop over all attributes and check if they should be created, renamed or imputed
        for attribute in self.attributes.values():
            df_log = DataStructure.create_attribute_columns(df_log, attribute)
            df_log = DataStructure.replace_with_nan(df_log, attribute)
            if len(attribute.na_rep_columns) > 0:  # impute values in case of missing values
                df_log = DataStructure.replace_nan_values_based_on_na_rep_columns(df_log, attribute)
            if attribute.na_rep_value is not None:
                df_log = DataStructure.replace_nan_values_based_on_na_rep_value(df_log, attribute)

            df_log = DataStructure.combine_attribute_columns(df_log, attribute)

        return df_log

    def split_df_log_into_combined_events(self, df_log: DataFrame):
        df_log["idx"] = df_log.reset_index().index
        if "timestamp" in self.attributes:
            raise ImportError(
                "Combined events cannot be split since an attribute with name timestamp is already defined")

        if "startTimestamp" in self.attributes and "completeTimestamp" in self.attributes:
            df_log_start = df_log.drop(columns=["completeTimestamp"])
            df_log_start["lifecycle"] = "start"
            df_log_start = df_log_start.rename(columns={"startTimestamp": "timestamp"})

            df_log_end = df_log.drop(columns=["startTimestamp"])
            df_log_end["lifecycle"] = "complete"
            df_log_end = df_log_end.rename(columns={"completeTimestamp": "timestamp"})

            df_log = pd.concat([df_log_start, df_log_end])
        elif "startTimestamp" in self.attributes:
            df_log["lifecycle"] = "start"
            df_log = df_log.rename(columns={"startTimestamp": "timestamp"})
        elif "completeTimestamp" in self.attributes:
            df_log["lifecycle"] = "complete"
            df_log = df_log.rename(columns={"completeTimestamp": "timestamp"})

        df_log = df_log.sort_values(by=["timestamp", "idx"])
        df_log = df_log.drop(columns=["idx"])
        return df_log

    def update_attributes(self):
        if "startTimestamp" in self.attributes and "completeTimestamp" in self.attributes:
            start_dt_format = self.get_datetime_formats()["startTimestamp"].format
            complete_dt_format = self.get_datetime_formats()["completeTimestamp"].format
            if start_dt_format != complete_dt_format:
                raise ValueError("startTimestamp and completeTimestamp have a different format")

            start_attribute = self.attributes["startTimestamp"]
            start_attribute.name = "timestamp"
            self.attributes["timestamp"] = start_attribute
            del self.attributes["startTimestamp"]
            del self.attributes["completeTimestamp"]


        elif "startTimestamp" in self.attributes:
            start_attribute = self.attributes["startTimestamp"]
            start_attribute.name = "timestamp"

            self.attributes["timestamp"] = start_attribute
            del self.attributes["startTimestamp"]
        elif "completeTimestamp" in self.attributes:
            complete_attribute = self.attributes["completeTimestamp"]
            complete_attribute.name = "timestamp"

            self.attributes["timestamp"] = complete_attribute
            del self.attributes["completeTimestamp"]

    def check_if_required_attributes_are_present(self, record_constructor):
        # loop over all required attributes of the record constructor
        # --> check whether they are required by self
        required = True
        for required_attribute in record_constructor.required_attributes:
            if required_attribute == "index":  # records of an always have an index, so records will have
                # the required attributes = index
                continue
            if required_attribute not in self.attributes or self.attributes[required_attribute].optional:
                # if the required attribute does not occur in the self.attributes or the attribute is
                # optional, then the record_constructor labels are not required
                required = False
        return required

    def determine_required_labels(self, records: List["RecordConstructor"]):
        # determine the required labels for the records defined by this datastructure
        # therefore, we have to check for all record constructors in records
        # whether the required attributes are present by the dataset description definition

        # initially, self.required_labels might be None, hence we need to make it an empty list.
        if self.required_labels is None:
            self.required_labels = []

        # loop over all record_constructors in records
        for record_constructor in records:
            # check whether self.labels appear in the record constructor
            if self.labels_appear_in_record_constructor(record_constructor):
                if record_constructor.prevalent_record.where_condition != "":
                    # a where condition is specified, then not all records require the record_constructor labels
                    required = False
                else:
                    required = self.check_if_required_attributes_are_present(record_constructor)
                if required:
                    # if required = true, then the record_constructor labels are required labels of self
                    self.required_labels.extend(record_constructor.record_labels)
        self.required_labels = list(set(self.required_labels))

    def get_required_labels(self, records: List["RecordConstructor"]):
        if self.required_labels is None:
            self.determine_required_labels(records)
        return self.required_labels

    def get_required_labels_str(self, records: List["RecordConstructor"]):
        """Method to convert the required labels determined by records to a string seperated by :
        :param records: the required
        """
        if self.required_labels is None:
            self.determine_required_labels(records)
        required_labels_w_records = self.required_labels[:]
        required_labels_w_records.insert(0, "Record")
        required_labels_str = ":".join(required_labels_w_records)
        return required_labels_str

    def read_data_set(self, file_name, use_sample, store_preprocessed_file=True,
                      use_preprocessed_file=False):
        preprocessed_file_name = self._get_preprocessed_file_name(file_name, use_sample)
        df_log = None
        if use_preprocessed_file:
            df_log = self.read_preprocessed_df_log(preprocessed_file_name)
        if df_log is None:  # no preprocessed file was read
            df_log = self.read_df_log(file_name, use_sample)
            if store_preprocessed_file:
                self.store_df_log(df_log, preprocessed_file_name)

        return df_log

    def read_df_log(self, file_name, use_sample):
        df_log = self.prepare_event_data_sets(file_name, use_sample)
        return df_log

    def read_preprocessed_df_log(self, preprocessed_file_name):
        preprocessed_file_path = os.path.join(self.preprocessed_file_directory, preprocessed_file_name)
        # first try to read preprocessed file
        if not os.path.exists(preprocessed_file_path):
            warning_message = f"No preprocessed file {preprocessed_file_name} found, preprocessed the file " \
                              f"instead"
            warnings.warn(warning_message)
            return None
        else:  # use_preprocessed_file and file exists
            df_log = pd.read_pickle(preprocessed_file_path)
        return df_log

    def store_df_log(self, df_log, preprocessed_file_name):
        # only store file if we did not read from the stored file
        os.makedirs(self.preprocessed_file_directory, exist_ok=True)

        preprocessed_file_path = os.path.join(self.preprocessed_file_directory, preprocessed_file_name)
        df_log.to_pickle(preprocessed_file_path)

    def _get_preprocessed_file_name(self, file_name, use_sample):
        # change extension from csv to pkl and add sample in case of sample
        sample_is_used = use_sample and len(self.samples) > 0
        preprocessed_file_name = f"{file_name[:-4]}_sample.pkl" if sample_is_used else f"{file_name[:-4]}.pkl"
        return preprocessed_file_name

    @staticmethod
    def create_record_id_column(df_log, file_name):
        record_id_column = df_log.index
        record_id_column = record_id_column.astype(str) + "_" + file_name[:-4]
        return record_id_column

    def prepare_event_data_sets(self, file_name, use_sample):
        dtypes = self.get_dtype_dict()
        required_columns = self.get_required_columns()

        true_values = self.true_values
        false_values = self.false_values

        if file_name.endswith('.csv'):
            df_log: DataFrame = pd.read_csv(os.path.join(self.file_directory, file_name), keep_default_na=True,
                                            usecols=required_columns, dtype=dtypes, true_values=true_values,
                                            false_values=false_values, sep=self.seperator, decimal=self.decimal,
                                            encoding=self.encoding)
        else:
            raise TypeError(f"The file extension of {file_name} is not implemented. Use .csv.")

        if use_sample and self.has_datetime_attribute():
            df_log = self.create_sample(file_name, df_log)

        df_log = self.preprocess_according_to_attributes(df_log)

        # all columns have been renamed to or constructed with the name attribute,
        # hence only keep those that match with a name attribute
        required_attributes = list([f"{attribute_name}_attribute" for attribute_name in self.attributes.keys()])
        required_attributes_mapping = {f"{attribute_name}_attribute": f"{attribute_name}" for attribute_name in
                                       self.attributes.keys()}
        df_log = df_log[required_attributes]
        df_log = df_log.rename(columns=required_attributes_mapping)

        if self.split_combined_events:
            df_log = self.split_df_log_into_combined_events(df_log)
            self.update_attributes()

        if self.add_log:
            df_log["log"] = file_name

        if self.add_index:
            df_log["index"] = df_log.index

        # drop all columns with only nan values
        df_log = df_log.dropna(how='all', axis=1)  # drop all columns in which all values are nan (empty)
        df_log = df_log.dropna(how='all')  # drop all columns in which all values are nan (empty)

        df_log["recordId"] = self.create_record_id_column(df_log, file_name)

        return df_log

    def determine_optional_labels_in_log(self, df_log, records):
        df_log["labels"] = ""

        if self.required_labels is None:
            self.determine_required_labels(records)

        if set(self.required_labels) == set(self.labels):
            return df_log

        for record_constructor in records:
            if self.labels_appear_in_record_constructor(record_constructor):
                if set(record_constructor.record_labels) <= set(self.required_labels):
                    continue  # record constructor only contains required labels

                # one of possible labels appear in record constructor
                should_have_label = self.all_required_attributes_are_present_in_df_log(df_log, record_constructor)
                if should_have_label.any():  # check whether there is still a row that can have the label
                    should_have_label = should_have_label & self.is_where_condition_satisfied(
                        df_log, record_constructor)
                if should_have_label.any():
                    for label in record_constructor.record_labels:
                        df_log.loc[should_have_label, "labels"] += ":" + label

        return df_log

    def labels_appear_in_record_constructor(self, record_constructor):
        if self.labels is not None:
            # check whether the labels have overlap with the labels of the record_constructor
            intersection = list(set(self.labels) & set(record_constructor.record_labels))
        else:
            # no labels are defined, hence we consider all labels of the record constructor
            intersection = record_constructor.record_labels

        return len(intersection) > 0

    def all_required_attributes_are_present_in_df_log(self, df_log, record_constructor):
        required_attributes_are_present = pd.Series(True, index=np.arange(len(df_log)), name='present')
        for required_attribute in record_constructor.required_attributes:
            if required_attribute == "index" or required_attribute == "log":
                continue
            if required_attribute in self.attributes:
                if self.attributes[required_attribute].optional:
                    # if the required attribute is optional in the structure, check whether the required attribute is
                    # not null
                    required_attributes_are_present = required_attributes_are_present & df_log[
                        required_attribute].notnull()
                # else the required attributes is required in the structure, hence nothing changes
            else:  # the required attribute is missing in the structure -> hence not all are present
                required_attributes_are_present = pd.Series(False, index=np.arange(len(df_log)), name='present')
                # since for all rows, at least an attribute is missing, we can return
                return required_attributes_are_present

        return required_attributes_are_present

    @staticmethod
    def is_where_condition_satisfied(df_log, record_constructor):
        # TODO split where_condition
        where_condition = record_constructor.prevalent_record.where_condition
        where_condition_satisfied = pd.Series(True, index=np.arange(len(df_log)), name='satisfied')
        if where_condition == "":
            return where_condition_satisfied
        where_conditions = where_condition.split("AND")
        for condition in where_conditions:
            stripped_condition = condition.strip()
            if "=" in condition:
                condition_list = stripped_condition.split("=")
                column_name = condition_list[0].strip()
                if "." in column_name:
                    column_name = column_name.split(".")[1].strip()
                column_value = condition_list[1].strip().strip('"')
                where_condition_satisfied = where_condition_satisfied & (
                        df_log[column_name].astype("string") == column_value)
            elif "STARTS WITH" in condition:
                condition_list = stripped_condition.split("STARTS WITH")
                column_name = condition_list[0].strip()
                if "." in column_name:
                    column_name = column_name.split(".")[1].strip()
                column_value = condition_list[1].strip().strip('"')
                where_condition_satisfied = where_condition_satisfied & df_log[
                    column_name].str.startswith(column_value)
            elif "ENDS WITH" in condition:
                condition_list = stripped_condition.split("ENDS WITH")
                column_name = condition_list[0].strip()
                if "." in column_name:
                    column_name = column_name.split(".")[1].strip()
                column_value = condition_list[1].strip().strip('"')
                where_condition_satisfied = where_condition_satisfied & df_log[
                    column_name].str.endswith(column_value)
        return where_condition_satisfied

    def get_datetime_formats(self) -> Dict[str, DatetimeObject]:
        datetime_formats = {}

        for attribute_name, attribute in self.attributes.items():
            if attribute.is_datetime:
                datetime_formats[attribute_name] = attribute.datetime_object

        return datetime_formats

    def get_attribute_value_pairs_filtered(self, exclude: bool) -> Dict[str, List[str]]:
        attribute_value_pairs = {}
        for attribute_name, attribute in self.attributes.items():
            if attribute.use_filter:
                attribute_value_pairs[attribute_name] \
                    = attribute.filter_exclude_values if exclude else attribute.filter_include_values

        return attribute_value_pairs


class DatasetDescriptions:
    def __init__(self, config: Configuration):
        path = config.dataset_description_path

        random.seed(1)
        with open(path, encoding='utf-8') as f:
            json_event_tables = json.load(f)

        self.structures = [DataStructure.from_dict(item) for item in json_event_tables]
        self.structures = [item for item in self.structures if item is not None]

    def get_structure_name_file_mapping(self):
        # request the file names per structure
        file_names = {}
        for structure in self.structures:
            file_names[structure.name] = structure.file_names
        return file_names

    def get_files_list(self):
        # request all file names specified in self.structures in a list
        files = []
        for structure in self.structures:
            files.extend(structure.file_names)
        return files



from dataclasses import dataclass
from enum import Enum


# from remote_authentication import remote
# if you want a remote connection, store credentials somewhere different (e.g. remote_authentication)

@dataclass
class Credentials:
    uri: str
    user: str
    password: str


# local credentials
local = Credentials(
    uri="bolt://localhost:7687",
    user="neo4j",
    password="12345678"
)

remote = None


class Connections(Enum):
    REMOTE = 1
    LOCAL = 2


connections_map = {
    Connections.REMOTE: remote,
    Connections.LOCAL: local
}
